import numpy as np
import pyarrow.compute as pc
import pytest

from ...coordinates.covariances import CoordinateCovariances
from ...coordinates.origin import Origin
from ...coordinates.spherical import SphericalCoordinates
from ...time import Timestamp
from ..associations import Associations
from ..detections import PointSourceDetections
from ..exposures import Exposures
from ..photometry import Photometry
from ..source_catalog import SourceCatalog


@pytest.fixture
def source_catalog() -> SourceCatalog:
    source_catalog = SourceCatalog.from_kwargs(
        id=["obs_01", "obs_02", "obs_03", "obs_04", "obs_05", "obs_06"],
        exposure_id=["exp_01", "exp_01", "exp_02", "exp_02", "exp_03", "exp_04"],
        time=Timestamp.from_kwargs(
            days=[59001, 59001, 59001, 59001, 59002, 59003],
            nanos=[0, 0.5 * 1e9, 20 * 1e9, 20 * 1e9, 0, 0],
            scale="utc",
        ),
        ra=[1, 2, 3, 4, 5, 6],
        ra_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        dec=[-1, -2, -3, -4, -5, -6],
        dec_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        radec_corr=[0.5, -0.5, 0.5, -0.5, 0.5, -0.5],
        mag=[10, 11, 12, 13, 14, 15],
        mag_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        fwhm=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        a=[0.2, 0.3, 0.4, 0.5, 0.6, 0.7],
        a_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        b=[0.2, 0.3, 0.4, 0.5, 0.6, 0.7],
        b_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        pa=[0, 30, 60, 90, 120, 150],
        pa_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        observatory_code=["X05", "X05", "I41", "I41", "X05", "I41"],
        filter=["g", "g", "r", "r", "i", "r"],
        exposure_start_time=Timestamp.from_kwargs(
            days=[59001, 59001, 59001, 59001, 59002, 59003],
            nanos=[0, 0, 0, 0, 0, 0],
            scale="utc",
        ),
        exposure_duration=[30, 30, 40, 40, 30, 40],
        exposure_seeing=[1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        exposure_depth_5sigma=[24.5, 24.5, 21.0, 21.0, 24.5, 21.5],
        object_id=["obj_01", "obj_02", "obj_03", "obj_04", "obj_05", "obj_06"],
        catalog_id=["Rubin", "Rubin", "ZTF", "ZTF", "Rubin", "ZTF"],
    )
    return source_catalog


def test_SourceCatalog_detections(source_catalog: SourceCatalog) -> None:
    # Test the detections method of the SourceCatalog class
    # returns the detections in the source catalog
    detections_actual = source_catalog.detections()
    detections_expected = PointSourceDetections.from_kwargs(
        id=["obs_01", "obs_02", "obs_03", "obs_04", "obs_05", "obs_06"],
        exposure_id=["exp_01", "exp_01", "exp_02", "exp_02", "exp_03", "exp_04"],
        time=Timestamp.from_kwargs(
            days=[59001, 59001, 59001, 59001, 59002, 59003],
            nanos=[0, 0.5 * 1e9, 20 * 1e9, 20 * 1e9, 0, 0],
            scale="utc",
        ),
        ra=[1, 2, 3, 4, 5, 6],
        ra_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        dec=[-1, -2, -3, -4, -5, -6],
        dec_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        mag=[10, 11, 12, 13, 14, 15],
        mag_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
    )

    assert detections_actual == detections_expected


def test_SourceCatalog_exposures(source_catalog: SourceCatalog) -> None:
    # Test the exposures method of the SourceCatalog class
    # returns the unique exposures in the source catalog
    exposures_actual = source_catalog.exposures()
    exposures_expected = Exposures.from_kwargs(
        id=["exp_01", "exp_02", "exp_03", "exp_04"],
        start_time=Timestamp.from_kwargs(
            days=[59001, 59001, 59002, 59003],
            nanos=[0, 0, 0, 0],
            scale="utc",
        ),
        duration=[30, 40, 30, 40],
        filter=["g", "r", "i", "r"],
        observatory_code=["X05", "I41", "X05", "I41"],
        seeing=[1.0, 1.0, 1.0, 1.0],
        depth_5sigma=[24.5, 21.0, 24.5, 21.5],
    )
    assert exposures_actual == exposures_expected


def test_SourceCatalog_associations(source_catalog: SourceCatalog) -> None:
    # Test the associations method of the SourceCatalog class
    # returns the associations in the source catalog
    associations_actual = source_catalog.associations()
    associations_expected = Associations.from_kwargs(
        detection_id=["obs_01", "obs_02", "obs_03", "obs_04", "obs_05", "obs_06"],
        object_id=["obj_01", "obj_02", "obj_03", "obj_04", "obj_05", "obj_06"],
    )
    assert associations_actual == associations_expected


def test_SourceCatalog_photometry(source_catalog: SourceCatalog) -> None:
    # Test the photometry method of the SourceCatalog class
    # returns the photometry in the source catalog
    photometry_actual = source_catalog.photometry()
    photometry_expected = Photometry.from_kwargs(
        time=Timestamp.from_kwargs(
            days=[59001, 59001, 59001, 59001, 59002, 59003],
            nanos=[0, 0.5 * 1e9, 20 * 1e9, 20 * 1e9, 0, 0],
            scale="utc",
        ),
        mag=[10, 11, 12, 13, 14, 15],
        mag_sigma=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
        filter=["g", "g", "r", "r", "i", "r"],
    )
    assert photometry_actual == photometry_expected


def test_SourceCatalog_coordinates(source_catalog: SourceCatalog) -> None:
    # Test the coordinates method of the SourceCatalog class
    # returns the astrometry in the source catalog as SphericalCoordinates
    coordinates_actual = source_catalog.coordinates()

    covariances = np.empty((len(source_catalog), 6, 6))
    covariances.fill(np.nan)

    ra_sigma = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6]) / 3600.0
    dec_sigma = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6]) / 3600.0
    radec_corr = np.array([0.5, -0.5, 0.5, -0.5, 0.5, -0.5])

    covariances[:, 1, 1] = ra_sigma**2
    covariances[:, 2, 2] = dec_sigma**2
    covariances[:, 1, 2] = radec_corr * ra_sigma * dec_sigma
    covariances[:, 2, 1] = covariances[:, 1, 2]

    coordinates_expected = SphericalCoordinates.from_kwargs(
        lon=[1, 2, 3, 4, 5, 6],
        lat=[-1, -2, -3, -4, -5, -6],
        time=Timestamp.from_kwargs(
            days=[59001, 59001, 59001, 59001, 59002, 59003],
            nanos=[0, 0.5 * 1e9, 20 * 1e9, 20 * 1e9, 0, 0],
            scale="utc",
        ),
        covariance=CoordinateCovariances.from_matrix(covariances),
        origin=Origin.from_kwargs(code=["X05", "X05", "I41", "I41", "X05", "I41"]),
        frame="equatorial",
    )

    assert (
        coordinates_actual.flattened_table()
        .drop_columns(["covariance.values"])
        .equals(
            coordinates_expected.flattened_table().drop_columns(["covariance.values"])
        )
    )
    np.testing.assert_almost_equal(
        coordinates_actual.covariance.to_matrix(),
        coordinates_expected.covariance.to_matrix(),
        decimal=12,
    )


def test_SourceCatalog_observers(source_catalog: SourceCatalog) -> None:
    # Test the observers method of the SourceCatalog class
    # returns the observer location at each detection time
    # correctly
    observers_actual = source_catalog.observers(exposure_midpoint=False)
    assert len(observers_actual) == 6
    assert (
        len(
            observers_actual.drop_duplicates(
                subset=["code", "coordinates.time.days", "coordinates.time.nanos"]
            )
        )
        == 5
    )
    assert pc.all(
        pc.equal(
            observers_actual.coordinates.time.days,
            [59001, 59001, 59001, 59001, 59002, 59003],
        )
    ).as_py()
    assert pc.all(
        pc.equal(
            observers_actual.coordinates.time.nanos,
            [0, 0.5 * 1e9, 20 * 1e9, 20 * 1e9, 0, 0],
        )
    ).as_py()
    assert pc.all(
        pc.equal(observers_actual.code, ["X05", "X05", "I41", "I41", "X05", "I41"])
    ).as_py()

    observers_actual = source_catalog.observers(exposure_midpoint=True)
    assert len(observers_actual) == 6
    assert (
        len(
            observers_actual.drop_duplicates(
                subset=["code", "coordinates.time.days", "coordinates.time.nanos"]
            )
        )
        == 4
    )
    assert pc.all(
        pc.equal(
            observers_actual.coordinates.time.days,
            [59001, 59001, 59001, 59001, 59002, 59003],
        )
    ).as_py()
    assert pc.all(
        pc.equal(
            observers_actual.coordinates.time.nanos,
            [15 * 1e9, 15 * 1e9, 20 * 1e9, 20 * 1e9, 15 * 1e9, 20 * 1e9],
        )
    ).as_py()
    assert pc.all(
        pc.equal(observers_actual.code, ["X05", "X05", "I41", "I41", "X05", "I41"])
    ).as_py()


def test_SourceCatalog_healpixels(source_catalog: SourceCatalog) -> None:
    # Test the healpixels method of the SourceCatalog class
    # returns the healpixels for each observation
    healpixels_actual = source_catalog.healpixels(nside=16)
    healpixels_expected = np.array([1087, 1087, 1085, 1085, 1085, 1079])
    np.testing.assert_array_equal(healpixels_actual, healpixels_expected)

    healpixels_actual = source_catalog.healpixels(nside=2)
    healpixels_expected = np.array([16, 16, 16, 16, 16, 16])
    np.testing.assert_array_equal(healpixels_actual, healpixels_expected)



import uuid
from typing import List, Literal, Optional, Tuple

import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.residuals import Residuals
from ..orbits.orbits import Orbits


def assign_duplicate_observations(
    orbits: "FittedOrbits", orbit_members: "FittedOrbitMembers"
) -> Tuple["FittedOrbits", "FittedOrbitMembers"]:
    """
    Assigns observations that have been assigned to multiple orbits to the orbit with the
    most observations, longest arc length, and lowest reduced chi2.

    Parameters
    ----------
    orbit_members : `~thor.orbit_determination.FittedOrbitMembers`
        Fitted orbit members.

    Returns
    -------
    filtered : `~thor.orbit_determination.FittedOrbits`
        Fitted orbits with duplicate assignments removed.
    filtered_orbit_members : `~thor.orbit_determination.FittedOrbitMembers`
        Fitted orbit members with duplicate assignments removed.
    """
    # Sorting by priority criteria
    orbits = orbits.sort_by(
        [
            ("num_obs", "descending"),
            ("arc_length", "descending"),
            ("reduced_chi2", "ascending"),
        ]
    )

    # Extracting unique observation IDs
    unique_obs_ids = pc.unique(orbit_members.column("obs_id"))

    # Dictionary to store the best orbit for each observation
    best_orbit_for_obs = {}

    # Iterate over each unique observation ID
    for obs_id in unique_obs_ids:
        # Filter orbit_members for the current observation ID
        mask = pc.equal(orbit_members.column("obs_id"), obs_id)
        current_obs_members = orbit_members.where(mask)

        # Extract orbit IDs that this observation belongs to
        obs_orbit_ids = current_obs_members.column("orbit_id")

        # Find the best orbit for this observation based on the criteria
        for sorted_orbit_id in orbits.column("orbit_id"):
            if pc.any(pc.is_in(sorted_orbit_id, value_set=obs_orbit_ids)).as_py():
                best_orbit_for_obs[obs_id.as_py()] = sorted_orbit_id.as_py()
                break

    # Iteratively update orbit_members to drop rows where obs_id is the same,
    # but orbit_id is not the best orbit_id for that observation
    for obs_id, best_orbit_id in best_orbit_for_obs.items():
        mask_to_remove = pc.and_(
            pc.equal(orbit_members.column("obs_id"), pa.scalar(obs_id)),
            pc.not_equal(orbit_members.column("orbit_id"), pa.scalar(best_orbit_id)),
        )
        orbit_members = orbit_members.apply_mask(pc.invert(mask_to_remove))

    # Filtering self based on the filtered orbit_members
    orbits_mask = pc.is_in(
        orbits.column("orbit_id"), value_set=orbit_members.column("orbit_id")
    )
    filtered_orbits = orbits.apply_mask(orbits_mask)

    return filtered_orbits, orbit_members


def drop_duplicate_orbits(
    orbits: "FittedOrbits",
    orbit_members: "FittedOrbitMembers",
    subset: Optional[List[str]] = None,
    keep: Literal["first", "last"] = "first",
) -> Tuple["FittedOrbits", "FittedOrbitMembers"]:
    """
    Drop duplicate orbits from the fitted orbits and remove
    the corresponding orbit members.

    Parameters
    ----------
    orbits : `~thor.orbit_determination.FittedOrbits`
        Fitted orbits.
    orbit_members : `~thor.orbit_determination.FittedOrbitMembers`
        Fitted orbit members.
    subset : list of str, optional
        Subset of columns to consider when dropping duplicates. If not specified all the columns
        specifying unique state are used: time, x, y, z, vx, vy, vz.
    keep : {'first', 'last'}, default 'first'
        If there are duplicate rows then keep the first or last row.

    Returns
    -------
    filtered : `~thor.orbit_determination.FittedOrbits`
        Fitted orbits without duplicates.
    filtered_orbit_members : `~thor.orbit_determination.FittedOrbitMembers`
        Fitted orbit members without duplicates.
    """
    if subset is None:
        subset = [
            "coordinates.time.days",
            "coordinates.time.nanos",
            "coordinates.x",
            "coordinates.y",
            "coordinates.z",
            "coordinates.vx",
            "coordinates.vy",
            "coordinates.vz",
        ]

    filtered = orbits.drop_duplicates(subset=subset, keep=keep)
    filtered_orbit_members = orbit_members.apply_mask(
        pc.is_in(orbit_members.orbit_id, filtered.orbit_id)
    )
    return filtered, filtered_orbit_members


class FittedOrbits(qv.Table):

    orbit_id = qv.LargeStringColumn(default=lambda: uuid.uuid4().hex)
    object_id = qv.LargeStringColumn(nullable=True)
    coordinates = CartesianCoordinates.as_column()
    arc_length = qv.Float64Column()
    num_obs = qv.Int64Column()
    chi2 = qv.Float64Column()
    reduced_chi2 = qv.Float64Column()
    iterations = qv.Int64Column(nullable=True)
    success = qv.BooleanColumn(nullable=True)
    status_code = qv.Int64Column(nullable=True)

    def to_orbits(self) -> Orbits:
        """
        Convert fitted orbits to orbits that can be used by
        a Propagator.

        Returns
        -------
        orbits : `~adam_core.orbits.Orbits`
            Orbits.
        """
        return Orbits.from_kwargs(
            orbit_id=self.orbit_id,
            object_id=self.object_id,
            coordinates=self.coordinates,
        )


class FittedOrbitMembers(qv.Table):

    orbit_id = qv.LargeStringColumn()
    obs_id = qv.LargeStringColumn()
    residuals = Residuals.as_column(nullable=True)
    solution = qv.BooleanColumn(nullable=True)
    outlier = qv.BooleanColumn(nullable=True)



"""
Runtime APIs for the bandpass-driven photometry implementation.

Bandpass curves are vendored from the SVO Filter Profile Service. Please see
`REFERENCES.md` for the required acknowledgement and citations when using this
service.
"""

from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
from importlib.resources import files
from typing import Final, Iterable

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq

from .tables import (
    AsteroidTemplates,
    BandpassCurves,
    ObservatoryBandMap,
    TemplateBandpassIntegrals,
)

_DATA_DIR = files("adam_core.photometry.bandpasses").joinpath("data")

_BANDPASS_CURVES_FILE: Final[str] = "bandpass_curves.parquet"
_OBS_BAND_MAP_FILE: Final[str] = "observatory_band_map.parquet"
_TEMPLATES_FILE: Final[str] = "asteroid_templates.parquet"
_INTEGRALS_FILE: Final[str] = "template_bandpass_integrals.parquet"
_SOLAR_SPECTRUM_FILE: Final[str] = "solar_spectrum.parquet"


def _to_string_array(
    values: pa.Array | pa.ChunkedArray | npt.NDArray[np.object_] | Iterable[str],
) -> pa.Array | pa.ChunkedArray:
    if isinstance(values, (pa.Array, pa.ChunkedArray)):
        return pc.cast(values, pa.large_string())
    arr = pa.array(list(values), type=pa.large_string())
    return arr


def _key_from_code_band(
    code: pa.Array | pa.ChunkedArray, band: pa.Array | pa.ChunkedArray
) -> pa.Array | pa.ChunkedArray:
    # Avoid Python loops; Arrow string concatenation is fast.
    sep = pa.scalar("|", type=pa.large_string())
    return pc.binary_join_element_wise(code, band, sep)


def _normalize_reported_band_for_station(
    codes: pa.Array | pa.ChunkedArray,
    bands: pa.Array | pa.ChunkedArray,
    *,
    only_if_code: str,
) -> pa.Array | pa.ChunkedArray:
    """
    Normalize station-specific reported-band encodings to the "reported band" strings
    expected by `ObservatoryBandMap`.

    Currently this is only needed for LSST (X05), where MPC/ADES encodings can include
    'Lg'/'Lr'/... and 'LSST_g'/... and sometimes 'Y' for y-band.

    Notes
    -----
    - This must be applied *after* canonical `filter_id` pass-through checks.
    - This should be lightweight and Arrow-native (no Python loops).
    """
    mask = pc.equal(codes, pa.scalar(str(only_if_code), type=pa.large_string()))
    if not bool(pc.any(mask).as_py()):
        return bands

    b = pc.cast(bands, pa.large_string())
    b = pc.utf8_trim_whitespace(b)

    # Strip "LSST_" prefix: LSST_g -> g
    has_lsst_prefix = pc.match_substring_regex(b, "^LSST_")
    b_lsst = pc.utf8_slice_codeunits(b, 5)
    b = pc.if_else(has_lsst_prefix, b_lsst, b)

    # Strip leading 'L' for 2-character encodings: Lg -> g, Ly -> y, etc.
    is_Lx = pc.match_substring_regex(b, "^L[ugrizy]$")
    b_Lx = pc.utf8_slice_codeunits(b, 1)
    b = pc.if_else(is_Lx, b_Lx, b)

    # Accept uppercase variants too: LY -> y, etc.
    is_LX = pc.match_substring_regex(b, "^L[UGRIZY]$")
    b_LX = pc.utf8_lower(pc.utf8_slice_codeunits(b, 1))
    b = pc.if_else(is_LX, b_LX, b)

    # Normalize Y -> y (both as raw 'Y' and as 'LY' after stripping doesn't apply).
    b = pc.if_else(pc.equal(b, "Y"), "y", b)

    # Apply only for the targeted station code.
    return pc.if_else(mask, b, bands)


@lru_cache(maxsize=1)
def load_bandpass_curves() -> BandpassCurves:
    path = _DATA_DIR.joinpath(_BANDPASS_CURVES_FILE)
    return BandpassCurves.from_parquet(path)


@lru_cache(maxsize=1)
def load_observatory_band_map() -> ObservatoryBandMap:
    path = _DATA_DIR.joinpath(_OBS_BAND_MAP_FILE)
    return ObservatoryBandMap.from_parquet(path)


@lru_cache(maxsize=1)
def load_asteroid_templates() -> AsteroidTemplates:
    path = _DATA_DIR.joinpath(_TEMPLATES_FILE)
    return AsteroidTemplates.from_parquet(path)


@lru_cache(maxsize=1)
def load_template_integrals() -> TemplateBandpassIntegrals:
    path = _DATA_DIR.joinpath(_INTEGRALS_FILE)
    return TemplateBandpassIntegrals.from_parquet(path)


@lru_cache(maxsize=1)
def _load_solar_spectrum() -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Load the adopted solar spectrum used for all precomputed integrals.

    Returns
    -------
    wavelength_nm, flux : ndarray
        Flux is arbitrary units suitable for relative photometry.
    """
    path = _DATA_DIR.joinpath(_SOLAR_SPECTRUM_FILE)
    table = pq.read_table(path)
    wl = np.asarray(
        table.column("wavelength_nm").to_numpy(zero_copy_only=False), dtype=np.float64
    )
    flux = np.asarray(
        table.column("flux").to_numpy(zero_copy_only=False), dtype=np.float64
    )
    return wl, flux


def _compute_filter_solar_norm_photon(
    filter_wavelength_nm: npt.NDArray[np.float64],
    filter_throughput: npt.NDArray[np.float64],
) -> float:
    """
    Compute the solar-weighted photon-counting normalization for a filter:

        D = ∫ F_sun(λ) * T(λ) * λ dλ

    This is used to convert our precomputed template×filter photon integrals
    (which are proportional to total photon counts) into a *band-averaged* quantity:

        <R>_filter = (∫ F_sun(λ) * R_ast(λ) * T(λ) * λ dλ) / D

    Using <R> makes V→filter color terms depend on the *shape* of the bandpass and
    the reflectance spectrum, rather than on absolute throughput scaling or filter width.
    """
    solar_wl, solar_flux = _load_solar_spectrum()

    wl = np.asarray(filter_wavelength_nm, dtype=np.float64)
    thr = np.asarray(filter_throughput, dtype=np.float64)
    if wl.ndim != 1 or thr.ndim != 1 or len(wl) != len(thr):
        raise ValueError("filter arrays must be 1D and have the same length")
    if len(wl) < 2:
        return float("nan")

    wl_min = max(float(solar_wl.min()), float(wl.min()))
    wl_max = min(float(solar_wl.max()), float(wl.max()))
    if wl_max <= wl_min:
        return float("nan")

    mask = (solar_wl >= wl_min) & (solar_wl <= wl_max)
    grid = solar_wl[mask]
    sun = solar_flux[mask]
    if len(grid) < 2:
        return float("nan")

    t = _interp_to_grid(wl, thr, grid)
    return float(np.trapezoid(sun * t * grid, grid))


@lru_cache(maxsize=1)
def _solar_norm_by_filter_id() -> dict[str, float]:
    """
    Cache solar-weighted throughput normalizations for all vendored filters.
    """
    curves = load_bandpass_curves()
    out: dict[str, float] = {}
    for fid, wl_list, thr_list in zip(
        curves.filter_id.to_pylist(),
        curves.wavelength_nm.to_pylist(),
        curves.throughput.to_pylist(),
    ):
        wl = np.asarray(wl_list, dtype=np.float64)
        thr = np.asarray(thr_list, dtype=np.float64)
        d = _compute_filter_solar_norm_photon(wl, thr)
        if not np.isfinite(d) or d <= 0.0:
            raise ValueError(f"Invalid solar normalization for filter_id '{fid}'")
        out[str(fid)] = float(d)
    return out


def _solar_norm_for_filter_ids(
    filter_ids: npt.NDArray[np.object_],
) -> npt.NDArray[np.float64]:
    norms = _solar_norm_by_filter_id()
    filt = np.asarray(filter_ids, dtype=object)
    out = np.empty(len(filt), dtype=np.float64)
    for i, fid in enumerate(filt.tolist()):
        d = norms.get(str(fid))
        if d is None:
            raise ValueError(f"Unknown filter_id for solar normalization: {fid}")
        out[i] = float(d)
    return out


def map_to_canonical_filter_bands(
    observatory_codes: (
        pa.Array | pa.ChunkedArray | npt.NDArray[np.object_] | Iterable[str]
    ),
    bands: pa.Array | pa.ChunkedArray | npt.NDArray[np.object_] | Iterable[str],
    *,
    allow_fallback_filters: bool = True,
) -> npt.NDArray[np.object_]:
    """
    Suggest canonical (vendored) bandpass filter IDs for a set of observations.

    This function is intended to be called by users *before* passing filters into
    bandpass-driven magnitude APIs. It applies the following resolution strategy:

    - If the provided band is already a canonical vendored `filter_id` (i.e., we have a
      curve for it), it is passed through unchanged.
    - Else, if (observatory_code, band) is present in `ObservatoryBandMap`, that mapping
      is used.
    - Else, apply a conservative fallback for generic bands:
        u/g/r/i/z -> SDSS_u/g/r/i/z
        y         -> PS1_y

    Parameters
    ----------
    observatory_codes : array-like
        MPC observatory codes.
    bands : array-like
        Reported band labels OR canonical vendored filter IDs.
    allow_fallback_filters : bool, optional
        If True, allow generic-band fallbacks when no (observatory_code, band) mapping is
        available:
          u/g/r/i/z -> SDSS_u/g/r/i/z
          y         -> PS1_y
        If False, raise if any row would require those fallbacks. Canonical `filter_id`
        inputs are always passed through. Defaults to True.

    Returns
    -------
    ndarray
        Canonical vendored `filter_id` strings.
    """
    codes = _to_string_array(observatory_codes)
    b = _to_string_array(bands)
    if len(codes) != len(b):
        raise ValueError(
            f"observatory_codes length ({len(codes)}) must match bands length ({len(b)})"
        )

    curves = load_bandpass_curves()
    mapping = load_observatory_band_map()

    # Pass-through if already a known filter_id.
    b_idx = pc.fill_null(pc.index_in(b, value_set=curves.filter_id), -1)
    is_known = np.asarray(b_idx.to_numpy(zero_copy_only=False), dtype=np.int32) >= 0

    out = np.empty(len(b), dtype=object)
    b_np = np.asarray(b.to_numpy(zero_copy_only=False), dtype=object)
    out[is_known] = b_np[is_known]

    # Try mapping table for the rest.
    need_map = ~is_known
    used_fallback = np.zeros(len(b), dtype=bool)
    if np.any(need_map):
        need_map_arr = pa.array(need_map.tolist(), type=pa.bool_())
        b_need = pc.if_else(need_map_arr, b, b)
        b_for_map = _normalize_reported_band_for_station(
            codes, b_need, only_if_code="X05"
        )

        keys = _key_from_code_band(codes, b_for_map)
        idx = pc.fill_null(pc.index_in(keys, value_set=mapping.key), -1)
        idx_np = np.asarray(idx.to_numpy(zero_copy_only=False), dtype=np.int32)

        mapped = need_map & (idx_np >= 0)
        if np.any(mapped):
            mapped_vals = mapping.filter_id.take(
                pa.array(idx_np[mapped], type=pa.int32())
            )
            out[mapped] = np.asarray(
                mapped_vals.to_numpy(zero_copy_only=False), dtype=object
            )

        missing = need_map & (idx_np < 0)
        if np.any(missing):
            missing_idx = np.nonzero(missing)[0]
            missing_bands = b_np[missing].astype(str)
            codes_np = np.asarray(
                codes.to_numpy(zero_copy_only=False), dtype=object
            ).astype(str)
            missing_codes = codes_np[missing].astype(str)

            fallback_map = {
                "u": "SDSS_u",
                "g": "SDSS_g",
                "r": "SDSS_r",
                "i": "SDSS_i",
                "z": "SDSS_z",
                "y": "PS1_y",
            }

            for j, code, band in zip(
                missing_idx.tolist(), missing_codes.tolist(), missing_bands.tolist()
            ):
                band_l = band.lower()
                fb = fallback_map.get(band_l)
                if fb is None:
                    continue
                out[j] = fb
                used_fallback[j] = True

    # Validate outputs are all present and correspond to known vendored curves.
    missing_out = np.nonzero(out == None)[0]  # noqa: E711
    if len(missing_out) > 0:
        codes_np = np.asarray(
            codes.to_numpy(zero_copy_only=False), dtype=object
        ).astype(str)
        unknown = [f"{codes_np[i]}|{str(b_np[i])}" for i in missing_out.tolist()]
        raise ValueError(
            "Unable to suggest canonical filter_id(s) for: " + ", ".join(unknown)
        )

    if (not allow_fallback_filters) and np.any(used_fallback):
        codes_np = np.asarray(
            codes.to_numpy(zero_copy_only=False), dtype=object
        ).astype(str)
        fallback_pairs = [
            f"{codes_np[i]}|{str(b_np[i])}"
            for i in np.nonzero(used_fallback)[0].tolist()
        ]
        raise ValueError(
            "No non-fallback mapping found for: "
            + ", ".join(fallback_pairs)
            + ". Set allow_fallback_filters=True to allow SDSS/PS1 fallbacks."
        )

    # Final guarantee: every output has a curve.
    out_arr = pa.array(out.tolist(), type=pa.large_string())
    out_idx = pc.fill_null(pc.index_in(out_arr, value_set=curves.filter_id), -1)
    out_idx_np = np.asarray(out_idx.to_numpy(zero_copy_only=False), dtype=np.int32)
    if np.any(out_idx_np < 0):
        bad = np.unique(
            np.asarray(out, dtype=object)[out_idx_np < 0].astype(str)
        ).tolist()
        raise ValueError(
            "Suggested filter_id(s) do not have vendored curves: " + repr(bad)
        )

    return np.asarray(out, dtype=object)


def assert_filter_ids_have_curves(
    filter_ids: pa.Array | pa.ChunkedArray | npt.NDArray[np.object_] | Iterable[str],
) -> None:
    """
    Raise if any `filter_id` is not present in vendored `BandpassCurves`.
    """
    curves = load_bandpass_curves()
    arr = _to_string_array(filter_ids)
    idx = pc.fill_null(pc.index_in(arr, value_set=curves.filter_id), -1)
    idx_np = np.asarray(idx.to_numpy(zero_copy_only=False), dtype=np.int32)
    if np.any(idx_np < 0):
        raw = np.asarray(arr.to_numpy(zero_copy_only=False), dtype=object)
        missing = np.unique(raw[idx_np < 0].astype(str)).tolist()
        raise ValueError(
            "Unknown filter_id(s) (no vendored bandpass curve): "
            + repr(missing)
            + ". Run map_to_canonical_filter_bands() first to map observatory bands to canonical filter_ids."
        )


def _get_integrals_precomputed(
    template_id: str, filter_ids: npt.NDArray[np.object_]
) -> npt.NDArray[np.float64]:
    """
    Return solar-weighted mean reflectance values for a known template_id across filter_ids.
    """
    filt = np.asarray(filter_ids, dtype=object)
    if filt.ndim != 1:
        raise ValueError("filter_ids must be a 1D array")
    if len(filt) == 0:
        return np.asarray([], dtype=np.float64)

    integrals = load_template_integrals()
    key = pa.array(
        [f"{template_id}|{str(x)}" for x in filt.tolist()], type=pa.large_string()
    )
    sep = pa.scalar("|", type=pa.large_string())
    value_set = pc.binary_join_element_wise(
        integrals.template_id, integrals.filter_id, sep
    )
    idx = pc.index_in(key, value_set=value_set)
    idx = pc.fill_null(idx, -1)
    idx_np = np.asarray(idx.to_numpy(zero_copy_only=False), dtype=np.int32)
    if np.any(idx_np < 0):
        missing = np.unique(
            np.asarray(key.to_numpy(zero_copy_only=False), dtype=object)[idx_np < 0]
        )
        raise ValueError(
            f"Missing precomputed integrals for template '{template_id}' and filters: {missing.tolist()}"
        )
    vals = integrals.integral_photon.take(pa.array(idx_np, type=pa.int32()))
    numer = np.asarray(vals.to_numpy(zero_copy_only=False), dtype=np.float64)
    denom = _solar_norm_for_filter_ids(filt)
    return numer / denom


def get_integrals(
    template_id: str, filter_ids: npt.NDArray[np.object_]
) -> npt.NDArray[np.float64]:
    """
    Return solar-weighted mean reflectance values for `template_id` across `filter_ids`.

    Supports both vendored templates (precomputed) and custom templates registered
    at runtime via `register_custom_template`.
    """
    tid = str(template_id)
    tmpl = _CUSTOM_TEMPLATES.get(tid)
    if tmpl is None:
        return _get_integrals_precomputed(tid, filter_ids)

    curves = load_bandpass_curves()
    filt = np.asarray(filter_ids, dtype=object)
    if filt.ndim != 1:
        raise ValueError("filter_ids must be a 1D array")

    denom = _solar_norm_for_filter_ids(filt)
    out = np.empty(len(filt), dtype=np.float64)
    for i, fid in enumerate(filt.tolist()):
        key = (tid, str(fid))
        cached = _CUSTOM_INTEGRALS.get(key)
        if cached is not None:
            out[i] = float(cached) / float(denom[i])
            continue
        mask = pc.equal(curves.filter_id, str(fid))
        sel = curves.apply_mask(mask)
        if len(sel) != 1:
            raise ValueError(
                f"Unknown filter_id '{fid}' for custom template integral computation"
            )
        wl = np.asarray(sel.wavelength_nm[0].as_py(), dtype=np.float64)
        thr = np.asarray(sel.throughput[0].as_py(), dtype=np.float64)
        val = _compute_integral_photon(wl, thr, tmpl.wavelength_nm, tmpl.reflectance)
        _CUSTOM_INTEGRALS[key] = val
        out[i] = float(val) / float(denom[i])
    return out


def compute_mix_integrals(
    weight_C: float,
    weight_S: float,
    filter_ids: npt.NDArray[np.object_],
) -> npt.NDArray[np.float64]:
    """
    Compute integrals for a C/S linear mix using precomputed base integrals.

    This avoids recomputing any convolution and supports arbitrary weights.
    """
    w_c = float(weight_C)
    w_s = float(weight_S)
    if not np.isfinite(w_c) or not np.isfinite(w_s):
        raise ValueError("weights must be finite")
    if w_c < 0.0 or w_s < 0.0:
        raise ValueError("weights must be non-negative")
    s = w_c + w_s
    if s <= 0.0:
        raise ValueError("at least one weight must be > 0")
    w_c /= s
    w_s /= s

    filt = np.asarray(filter_ids, dtype=object)
    ints_c = get_integrals("C", filt)
    ints_s = get_integrals("S", filt)
    return w_c * ints_c + w_s * ints_s


def bandpass_delta_mag(
    composition: str | tuple[float, float],
    source_filter_id: str,
    target_filter_id: str,
) -> float:
    """
    Compute a constant magnitude offset between two canonical filters for a composition.

    The delta is defined as:

        Δm = m_target - m_source = -2.5 log10(<R>_target / <R>_source)

    where <R> is the solar-weighted band-averaged reflectance for the composition
    (see `get_integrals`).
    """
    src = str(source_filter_id)
    tgt = str(target_filter_id)
    if not src or not tgt:
        raise ValueError("source_filter_id and target_filter_id must be non-empty")
    if src == tgt:
        return 0.0

    ids = np.asarray([src, tgt], dtype=object)
    if isinstance(composition, str):
        integrals = get_integrals(str(composition), ids)
    else:
        w_c, w_s = composition
        integrals = compute_mix_integrals(float(w_c), float(w_s), ids)

    i_src = float(integrals[0])
    i_tgt = float(integrals[1])
    if not np.isfinite(i_src) or not np.isfinite(i_tgt) or i_src <= 0.0 or i_tgt <= 0.0:
        raise ValueError(f"Invalid integrals for delta magnitude {src} -> {tgt}")
    return float(-2.5 * np.log10(i_tgt / i_src))


def bandpass_color_terms(
    composition: str | tuple[float, float],
    *,
    source_filter_id: str = "V",
    target_filter_ids: Iterable[str] | None = None,
) -> dict[str, float]:
    """
    Return delta magnitudes relative to `source_filter_id` for a set of canonical filters.

    Returns
    -------
    dict
        Mapping: target_filter_id -> Δm where Δm = m_target - m_source.
    """
    src = str(source_filter_id)
    if not src:
        raise ValueError("source_filter_id must be non-empty")

    if target_filter_ids is None:
        targets = [str(x) for x in load_bandpass_curves().filter_id.to_pylist()]
    else:
        targets = [str(x) for x in target_filter_ids]

    targets = [t for t in targets if t and t != src]
    if not targets:
        return {}

    ids = np.asarray([src] + targets, dtype=object)
    if isinstance(composition, str):
        integrals = get_integrals(str(composition), ids)
    else:
        w_c, w_s = composition
        integrals = compute_mix_integrals(float(w_c), float(w_s), ids)

    i_src = float(integrals[0])
    if not np.isfinite(i_src) or i_src <= 0.0:
        raise ValueError(f"Invalid integral for source filter '{src}'")

    deltas = -2.5 * np.log10(np.asarray(integrals[1:], dtype=np.float64) / i_src)
    return {t: float(d) for t, d in zip(targets, deltas)}


@dataclass(frozen=True)
class _CustomTemplate:
    wavelength_nm: npt.NDArray[np.float64]
    reflectance: npt.NDArray[np.float64]


_CUSTOM_TEMPLATES: dict[str, _CustomTemplate] = {}
_CUSTOM_INTEGRALS: dict[tuple[str, str], float] = {}


def register_custom_template(
    template_id: str,
    wavelength_nm: npt.NDArray[np.float64],
    reflectance: npt.NDArray[np.float64],
) -> None:
    """
    Register a custom reflectance template in-process and compute missing integrals lazily.

    Notes
    -----
    - This does not modify vendored Parquet data.
    - This is intended for NumPy paths. JAX-compiled tables in `magnitude.py`
      cannot be extended at runtime without recompilation.
    """
    tid = str(template_id)
    if not tid:
        raise ValueError("template_id must be non-empty")

    wl = np.asarray(wavelength_nm, dtype=np.float64)
    rf = np.asarray(reflectance, dtype=np.float64)
    if wl.ndim != 1 or rf.ndim != 1:
        raise ValueError("wavelength_nm and reflectance must be 1D arrays")
    if len(wl) != len(rf):
        raise ValueError("wavelength_nm and reflectance must have the same length")
    if len(wl) < 2:
        raise ValueError("template arrays must have at least 2 points")

    order = np.argsort(wl)
    wl = wl[order]
    rf = rf[order]
    if np.any(~np.isfinite(wl)) or np.any(~np.isfinite(rf)):
        raise ValueError("template arrays must be finite")
    if np.any(np.diff(wl) <= 0):
        raise ValueError("wavelength_nm must be strictly increasing")

    _CUSTOM_TEMPLATES[tid] = _CustomTemplate(wavelength_nm=wl, reflectance=rf)
    _clear_custom_cache()


def _interp_to_grid(
    x: npt.NDArray[np.float64],
    y: npt.NDArray[np.float64],
    x_new: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    return np.interp(x_new, x, y, left=0.0, right=0.0).astype(np.float64, copy=False)


def _compute_integral_photon(
    filter_wavelength_nm: npt.NDArray[np.float64],
    filter_throughput: npt.NDArray[np.float64],
    template_wavelength_nm: npt.NDArray[np.float64],
    template_reflectance: npt.NDArray[np.float64],
) -> float:
    solar_wl, solar_flux = _load_solar_spectrum()

    wl_min = max(
        float(solar_wl.min()),
        float(filter_wavelength_nm.min()),
        float(template_wavelength_nm.min()),
    )
    wl_max = min(
        float(solar_wl.max()),
        float(filter_wavelength_nm.max()),
        float(template_wavelength_nm.max()),
    )
    if wl_max <= wl_min:
        return float("nan")

    # Use the solar spectrum sampling as the shared grid (already dense and stable).
    mask = (solar_wl >= wl_min) & (solar_wl <= wl_max)
    wl = solar_wl[mask]
    sun = solar_flux[mask]
    if len(wl) < 2:
        return float("nan")

    t = _interp_to_grid(filter_wavelength_nm, filter_throughput, wl)
    r = _interp_to_grid(template_wavelength_nm, template_reflectance, wl)
    integrand = sun * r * t * wl
    return float(np.trapezoid(integrand, wl))


def _clear_custom_cache() -> None:
    _CUSTOM_INTEGRALS.clear()



from pathlib import Path

import numpy as np
import pytest

# Check if optional dependencies are available
try:
    from adam_assist import ASSISTPropagator

    HAS_ASSIST = True
except ImportError:
    HAS_ASSIST = False

try:
    import geopandas as gpd  # noqa: F401
    import plotly.graph_objects as go

    HAS_PLOTLY = True
except ImportError:
    HAS_PLOTLY = False

from adam_core.dynamics.impacts import CollisionEvent
from adam_core.dynamics.plots import (
    generate_impact_visualization_data,
    plot_impact_simulation,
    plot_risk_corridor,
)
from adam_core.orbits import Orbits, VariantOrbits
from adam_core.time import Timestamp

# Define test data directory - adjust as necessary
TEST_DATA_DIR = Path(__file__).parent / "data"


def load_test_data():
    """
    Load test data from parquet files.

    Returns
    -------
    tuple
        Orbit, VariantOrbits, and CollisionEvent objects loaded from parquet files.
    """
    # Check if test data files exist
    orbit_file = TEST_DATA_DIR / "orbits.parquet"
    variants_file = TEST_DATA_DIR / "variants.parquet"
    impacts_file = TEST_DATA_DIR / "collisions.parquet"

    # Check if all required test files exist
    files_exist = all(f.exists() for f in [orbit_file, variants_file, impacts_file])
    if not files_exist:
        pytest.skip("Test data files not found. Will be populated later by the user.")

    # Load the data from parquet files
    orbit = Orbits.from_parquet(orbit_file)
    variant_orbits = VariantOrbits.from_parquet(variants_file)
    impacts = CollisionEvent.from_parquet(impacts_file)

    return orbit, variant_orbits, impacts


@pytest.mark.skipif(not HAS_ASSIST, reason="ASSISTPropagator not available")
@pytest.mark.skipif(not HAS_PLOTLY, reason="Plotly not available")
@pytest.mark.parametrize(
    "time_step,time_range,max_processes",
    [
        (5, 60, None),  # Default values
        (10, 30, 1),  # Smaller range, single process
    ],
)
def test_generate_impact_visualization_data(time_step, time_range, max_processes):
    """
    Test the generate_impact_visualization_data function using data from parquet files.

    Parameters
    ----------
    time_step : float
        Time step for propagation in minutes.
    time_range : float
        Time range for propagation in minutes.
    max_processes : int or None
        Maximum number of processes to use for propagation.
    """
    # Load test data
    orbit, variant_orbits, impacts = load_test_data()

    # Create propagator
    propagator = ASSISTPropagator()

    # Call the function
    propagation_times, propagated_best_fit_orbit, propagated_variants = (
        generate_impact_visualization_data(
            orbit,
            variant_orbits,
            impacts,
            propagator,
            time_step=time_step,
            time_range=time_range,
            max_processes=max_processes,
        )
    )

    # Check types of returned objects
    assert isinstance(propagation_times, Timestamp)
    assert isinstance(propagated_best_fit_orbit, Orbits)
    assert isinstance(propagated_variants, dict)

    # Check that propagation times have reasonable length
    # based on time_step and time_range
    impact_times = impacts.coordinates.time.mjd().to_numpy(zero_copy_only=False)
    time_span_days = (np.max(impact_times) - np.min(impact_times)) + (
        time_range / 60 / 24 * 2
    )
    expected_num_times = int(time_span_days * 24 * 60 / time_step) + 1
    # Allow for some flexibility in the number of time steps due to rounding
    assert len(propagation_times) >= expected_num_times * 0.8
    assert len(propagation_times) <= expected_num_times * 1.2

    assert "Non-Impacting" in propagated_variants
    for impact_body in impacts.collision_object.code.unique().to_pylist():
        assert impact_body in propagated_variants

    # Check that propagated orbits have the expected time points
    assert len(propagated_best_fit_orbit) == len(propagation_times) * len(orbit)


@pytest.mark.skipif(not HAS_ASSIST, reason="ASSISTPropagator not available")
@pytest.mark.skipif(not HAS_PLOTLY, reason="Plotly not available")
@pytest.mark.parametrize(
    "time_step,time_range",
    [
        (5, 60),  # Default values
        (10, 30),  # Different values
    ],
)
def test_plot_impact_simulation(time_step, time_range):
    """
    Test the plot_impact_simulation function using data from parquet files.

    Parameters
    ----------
    time_step : float
        Time step for propagation in minutes.
    time_range : float
        Time range for propagation in minutes.
    """
    # Load test data
    orbit, variant_orbits, impacts = load_test_data()

    # Create propagator
    propagator = ASSISTPropagator()

    # Generate data for visualization
    propagation_times, propagated_best_fit_orbit, propagated_variants = (
        generate_impact_visualization_data(
            orbit,
            variant_orbits,
            impacts,
            propagator,
            time_step=time_step,
            time_range=time_range,
            max_processes=1,
        )
    )

    # Test with different parameter combinations
    test_cases = [
        {
            "title": "Test Impact Simulation",
            "logo": True,
            "sample_impactors": 0.5,
            "sample_non_impactors": 0.5,
        },
        {
            "title": None,
            "logo": False,
            "show_impacting": False,
            "show_non_impacting": True,
        },
        {
            "title": "Minimal Test",
            "grid": False,
            "show_earth": True,
            "show_moon": False,
        },
    ]

    for params in test_cases:
        fig = plot_impact_simulation(
            propagation_times,
            propagated_best_fit_orbit,
            propagated_variants,
            impacts,
            **params,
        )

        # Verify the figure is created correctly
        assert isinstance(fig, go.Figure)
        assert len(fig.frames) > 0

        # Check for expected elements based on params
        if params.get("show_earth", True):
            assert any(trace.name == "Earth" for trace in fig.data)

        if params.get("show_moon", True):
            assert any(trace.name == "Moon" for trace in fig.data)


@pytest.mark.skipif(not HAS_PLOTLY, reason="Plotly not available")
def test_plot_risk_corridor():
    """
    Test the plot_risk_corridor function using data from parquet files.
    """
    # Load test data
    _, _, impacts = load_test_data()

    # Test with different parameter combinations
    test_cases = [
        {"title": "Test Risk Corridor", "logo": True, "height": 600, "width": 800},
        {"title": None, "logo": False},
        {"title": "Custom Risk Corridor"},
    ]

    for params in test_cases:
        try:
            fig = plot_risk_corridor(impacts, **params)

            # Verify the figure is created correctly
            assert isinstance(fig, go.Figure)
            assert len(fig.frames) > 0

            # Check for expected title
            if params.get("title") is not None:
                assert params["title"] in fig.layout.title.text
            else:
                assert "Risk Corridor" in fig.layout.title.text

        except ValueError as e:
            # If no Earth impacts in the test data, this is expected
            if "No Earth impacts found" in str(e):
                pytest.skip(
                    "No Earth impacts in test data, skipping risk corridor test"
                )
            else:
                raise



import warnings
from typing import Union

import numpy as np
import numpy.typing as npt
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv
from mpc_obscodes import mpc_obscodes
from timezonefinder import TimezoneFinder
from typing_extensions import Self

from ..constants import Constants as c
from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.origin import OriginCodes
from ..time import Timestamp

R_EARTH_EQUATORIAL = c.R_EARTH_EQUATORIAL
R_EARTH_POLAR = c.R_EARTH_POLAR
E_EARTH = np.sqrt(1 - (R_EARTH_POLAR / R_EARTH_EQUATORIAL) ** 2)


class ObservatoryParallaxCoefficients(qv.Table):
    code = qv.LargeStringColumn()
    longitude = qv.Float64Column()
    cos_phi = qv.Float64Column()
    sin_phi = qv.Float64Column()
    name = qv.LargeStringColumn()

    def lon_lat(self) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """
        Return the longitude and latitude of the observatories in degrees.

        This is only valid for Earth-based observatories.

        Returns
        -------
        longitude : np.ndarray
            The longitude of the observatories in degrees. In the range -180 to 180 degrees,
            with positive values east of the prime meridian.
        latitude : np.ndarray
            The latitude of the observatories in degrees. In the range -90 to 90 degrees,
            with positive values north of the equator.
        """
        # Filter out Space-based observatories
        mask = pc.is_nan(self.longitude).to_numpy(zero_copy_only=False)

        longitude = np.where(
            mask, np.nan, self.longitude.to_numpy(zero_copy_only=False)
        )
        tan_phi_geo = np.where(
            mask,
            np.nan,
            self.sin_phi.to_numpy(zero_copy_only=False)
            / self.cos_phi.to_numpy(zero_copy_only=False),
        )
        latitude_geodetic = np.arctan(tan_phi_geo / (1 - E_EARTH**2))

        # Scale longitude to -180 to 180
        longitude = np.where(longitude > 180, longitude - 360, longitude)

        return longitude, np.degrees(latitude_geodetic)

    def timezone(self) -> npt.NDArray[np.str_]:
        """
        Return the timezone of the observatories in hours.

        Returns
        -------
        timezone : np.ndarray
            The timezone of the observatories in hours.
        """
        tf = TimezoneFinder()
        lon, lat = self.lon_lat()
        time_zones = np.array(
            [
                tz if not np.isnan(lon_i) else "None"
                for lon_i, lat_i in zip(lon, lat)
                for tz in [tf.timezone_at(lng=lon_i, lat=lat_i)]
            ]
        )
        return time_zones


# Read MPC extended observatory codes file
# Ignore warning about pandas deprecation
with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore",
        message="The behavior of 'to_datetime' with 'unit' when parsing strings is deprecated",
    )
    OBSCODES = pd.read_json(
        mpc_obscodes,
        orient="index",
        dtype={"Longitude": float, "cos": float, "sin": float, "Name": str},
        encoding_errors="strict",
        precise_float=True,
    )
    OBSCODES.reset_index(inplace=True, names=["code"])

OBSERVATORY_PARALLAX_COEFFICIENTS = ObservatoryParallaxCoefficients.from_kwargs(
    code=OBSCODES["code"].values,
    longitude=OBSCODES["Longitude"].values,
    cos_phi=OBSCODES["cos"].values,
    sin_phi=OBSCODES["sin"].values,
    name=OBSCODES["Name"].values,
)

OBSERVATORY_CODES = {
    x for x in OBSERVATORY_PARALLAX_COEFFICIENTS.code.to_numpy(zero_copy_only=False)
}


class Observers(qv.Table):
    code = qv.LargeStringColumn(nullable=False)
    coordinates = CartesianCoordinates.as_column()

    @classmethod
    def from_codes(
        cls, codes: Union[list, npt.NDArray[np.str_], pa.Array], times: Timestamp
    ) -> Self:
        """
        Create an Observers table from a list of codes and times. The codes and times
        do not need to be unique and are assumed to belong to each other in an element-wise fashion.
        The observer state will be calculated  correctly matched to the input times and
        replicated for duplicate times.

        Parameters
        ----------
        codes : Union[list, npt.NDArray[np.str], pa.Array] (N)
            MPC observatory codes for which to find the states.
        times : Timestamp (N)
            Epochs for which to find the observatory locations.

        Returns
        -------
        observers : `~adam_core.observers.observers.Observers` (N)
            The observer and its state at each time.
        """
        if len(codes) != len(times):
            raise ValueError("codes and times must have the same length.")

        if not isinstance(codes, pa.Array):
            codes = pa.array(codes, type=pa.large_string())

        class IndexedObservers(qv.Table):
            index = qv.UInt64Column()
            observers = Observers.as_column()

        indexed_observers = IndexedObservers.empty()

        # Loop through each unique code and calculate the observer's
        # state for each time (these can be non-unique as cls.from_code
        # will handle this)
        for code in pc.unique(codes):

            indices = pc.indices_nonzero(pc.equal(codes, code))
            times_code = times.take(indices)

            observers_i = cls.from_code(
                code.as_py(),
                times_code,
            )

            indexed_observers_i = IndexedObservers.from_kwargs(
                index=indices,
                observers=observers_i,
            )

            indexed_observers = qv.concatenate([indexed_observers, indexed_observers_i])
            if indexed_observers.fragmented():
                indexed_observers = qv.defragment(indexed_observers)

        return indexed_observers.sort_by("index").observers

    @classmethod
    def from_code(cls, code: Union[str, OriginCodes], times: Timestamp) -> Self:
        """
        Instantiate an Observers table with a single code and multiple times.
        Times do not need to be unique. The observer state will be calculated
        for each time and correctly matched to the input times and replicated for
        duplicate times.

        To load multiple codes, use `from_code` and then concatenate the tables.

        Note that NAIF origin codes may not be supported by `~adam_core.propagator.Propagator`
        classes such as PYOORB.

        Parameters
        ----------
        code : Union[str, OriginCodes]
            MPC observatory code or NAIF origin code for which to find the states.
        times : Timetamp (N)
            Epochs for which to find the observatory locations.

        Returns
        -------
        observers : `~adam_core.observers.observers.Observers` (N)
            The observer and its state at each time.

        Examples
        --------
        >>> import numpy as np
        >>> from adam_core.time import Timestamp
        >>> from adam_core.observers import Observers
        >>> times = Timestamp.from_mjd(np.arange(59000, 59000 + 100), scale="tdb")
        >>> observers = Observers.from_code("X05", times)
        """
        from .state import get_observer_state

        if isinstance(code, OriginCodes):
            code_str = code.name
        elif isinstance(code, str):
            code_str = code
        else:
            err = "Code should be a str or an `~adam_core.coordinates.origin.OriginCodes`."
            raise ValueError(err)

        return cls.from_kwargs(
            code=[code_str] * len(times),
            coordinates=get_observer_state(code, times),
        )

    def iterate_codes(self):
        """
        Iterate over the codes in the Observers table.

        Yields
        ------
        code : str
            The code for observer.
        observers : `~adam_core.observers.observers.Observers`
            The Observers table for this observer.
        """
        for code in self.code.unique().sort():
            yield code.as_py(), self.select("code", code)



import numpy as np
import pytest

from ...coordinates.cartesian import CartesianCoordinates
from ...coordinates.origin import Origin
from ...observers.observers import Observers
from ...time import Timestamp
from ..magnitude import (
    JAX_CHUNK_SIZE,
    calculate_apparent_magnitude_v,
    calculate_apparent_magnitude_v_and_phase_angle,
    calculate_phase_angle,
    convert_magnitude,
)


def _padded_size(n: int, pad_to: int) -> int:
    """Round n up to the next multiple of pad_to."""
    if pad_to <= 0:
        raise ValueError("pad_to must be > 0")
    return int(((n + pad_to - 1) // pad_to) * pad_to)


def _make_benchmark_case(n: int = 2048):
    rng = np.random.default_rng(123)
    time = Timestamp.from_mjd(np.full(n, 60000), scale="tdb")

    observer = Observers.from_kwargs(
        code=["500"] * n,
        coordinates=CartesianCoordinates.from_kwargs(
            x=np.full(n, 1.0),
            y=np.zeros(n),
            z=np.zeros(n),
            vx=np.zeros(n),
            vy=np.zeros(n),
            vz=np.zeros(n),
            time=time,
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"] * n),
        ),
    )

    obj = CartesianCoordinates.from_kwargs(
        x=rng.uniform(1.2, 3.0, size=n),
        y=rng.uniform(0.1, 2.0, size=n),
        z=rng.uniform(-0.5, 0.5, size=n),
        vx=np.zeros(n),
        vy=np.zeros(n),
        vz=np.zeros(n),
        time=time,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"] * n),
    )

    H = rng.uniform(10.0, 25.0, size=n)
    G = rng.uniform(0.0, 1.0, size=n)
    return H, obj, observer, G


def _make_convert_case(n: int = 262144):
    rng = np.random.default_rng(123)
    mags = rng.uniform(10.0, 25.0, size=n).astype(float)

    # Typical usage: many V-band magnitudes converted to per-observation filters.
    source = np.full(n, "V", dtype=object)
    targets = np.array(
        ["V", "SDSS_g", "SDSS_r", "SDSS_i", "LSST_g", "LSST_r", "DECam_r"],
        dtype=object,
    )
    target = rng.choice(targets, size=n).astype(object)
    return mags, source, target


@pytest.mark.parametrize(
    "n",
    [256, 16384, 65536, 262144, 1048576],
    ids=lambda x: f"n={x}",
)
@pytest.mark.benchmark(group="photometry_magnitude_apparent")
def test_benchmark_calculate_apparent_magnitude_jax(benchmark, n):
    pad_to = JAX_CHUNK_SIZE
    n_padded = _padded_size(n, pad_to)
    H, obj, observer, G = _make_benchmark_case(n=n_padded)

    # Warm-up to exclude JIT compilation from the benchmark measurement.
    _ = calculate_apparent_magnitude_v(H, obj, observer, G=G)

    def run():
        out = calculate_apparent_magnitude_v(H, obj, observer, G=G)
        return out

    out = benchmark(run)
    out = out[:n]  # window back to requested n
    assert len(out) == n


@pytest.mark.parametrize(
    "n",
    [256, 16384, 65536, 262144, 1048576],
    ids=lambda x: f"n={x}",
)
@pytest.mark.benchmark(group="photometry_phase_angle")
def test_benchmark_calculate_phase_angle_jax(benchmark, n):
    pad_to = JAX_CHUNK_SIZE
    n_padded = _padded_size(n, pad_to)
    _, obj, observer, _ = _make_benchmark_case(n=n_padded)

    # Warm-up to exclude JIT compilation from the benchmark measurement.
    _ = calculate_phase_angle(obj, observer)

    def run():
        out = calculate_phase_angle(obj, observer)
        return out

    out = benchmark(run)
    out = out[:n]
    assert len(out) == n


@pytest.mark.parametrize(
    "n",
    [256, 16384, 65536, 262144, 1048576],
    ids=lambda x: f"n={x}",
)
@pytest.mark.benchmark(group="photometry_magnitude_apparent_and_phase_angle")
def test_benchmark_calculate_apparent_magnitude_and_phase_jax(benchmark, n):
    pad_to = JAX_CHUNK_SIZE
    n_padded = _padded_size(n, pad_to)
    H, obj, observer, G = _make_benchmark_case(n=n_padded)

    # Warm-up to exclude JIT compilation from the benchmark measurement.
    _ = calculate_apparent_magnitude_v_and_phase_angle(H, obj, observer, G=G)

    def run():
        mags, alpha = calculate_apparent_magnitude_v_and_phase_angle(
            H, obj, observer, G=G
        )
        return mags, alpha

    mags, alpha = benchmark(run)
    mags = np.asarray(mags)[:n]
    alpha = np.asarray(alpha)[:n]
    assert len(mags) == n
    assert len(alpha) == n


@pytest.mark.parametrize(
    "n",
    [256, 16384, 65536, 262144, 1048576],
    ids=lambda x: f"n={x}",
)
@pytest.mark.benchmark(group="photometry_magnitude_convert_magnitude")
def test_benchmark_convert_magnitude_bandpass(benchmark, n):
    pad_to = JAX_CHUNK_SIZE
    n_padded = _padded_size(n, pad_to)
    mags, src, tgt = _make_convert_case(n=n_padded)

    def run():
        return convert_magnitude(mags, src, tgt, composition="NEO")

    out = benchmark(run)
    out = np.asarray(out)[:n]
    assert len(out) == n


@pytest.mark.benchmark(group="photometry_magnitude_convert_magnitude_shape_sweep")
def test_benchmark_convert_magnitude_bandpass_shape_sweep(benchmark):
    """
    Shape-sweep benchmark to ensure runtime does not exhibit repeated "first call" spikes
    across varying `n`.
    """
    pad_to = JAX_CHUNK_SIZE
    sizes = [256, 12345, 65536, 99999, 262144]
    cases = []
    for n in sizes:
        n_padded = _padded_size(n, pad_to)
        mags, src, tgt = _make_convert_case(n=n_padded)
        cases.append((mags, src, tgt))

    def run():
        out = None
        for mags, src, tgt in cases:
            out = convert_magnitude(mags, src, tgt, composition="NEO")
        return out

    out = benchmark(run)
    assert out is not None


def test_benchmark_calculate_apparent_magnitude_jax_shape_sweep(benchmark):
    pad_to = JAX_CHUNK_SIZE
    sizes = [256, 12345, 65536, 99999, 262144]
    cases = []
    for n in sizes:
        n_padded = _padded_size(n, pad_to)
        H, obj, observer, G = _make_benchmark_case(n=n_padded)
        cases.append((H, obj, observer, G))

    # Warm-up: compile once before benchmarking the sweep.
    H0, obj0, observer0, G0 = cases[0]
    _ = calculate_apparent_magnitude_v(H0, obj0, observer0, G=G0)

    def run():
        out = None
        for H, obj, observer, G in cases:
            out = calculate_apparent_magnitude_v(H, obj, observer, G=G)
        return out

    out = benchmark(run)
    assert out is not None



import logging
from typing import Dict, Optional

import numpy as np
import quivr as qv
import spiceypy as sp
from astropy import units as u
from naif_de440 import de440
from naif_earth_itrf93 import earth_itrf93
from naif_eop_high_prec import eop_high_prec
from naif_eop_historical import eop_historical
from naif_eop_predict import eop_predict
from naif_leapseconds import leapseconds

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.origin import OriginCodes
from ..coordinates.transform import transform_coordinates
from ..orbits import Orbits
from ..propagator import Propagator
from ..time import Timestamp
from ..utils.spice import setup_SPICE

DEFAULT_KERNELS = [
    leapseconds,
    de440,
    eop_predict,
    eop_historical,
    eop_high_prec,
    earth_itrf93,
]

# Add after DEFAULT_KERNELS
logger = logging.getLogger(__name__)

J2000_TDB_JD = 2451545.0


def fit_chebyshev(
    coordinates: CartesianCoordinates,
    window_start: float,
    window_end: float,
    degree: int,
) -> tuple[np.ndarray, float, float]:
    """
    Fit Chebyshev polynomials to position and velocity data.

    Parameters
    ----------
    coordinates : CartesianCoordinates
        Coordinates to fit
    window_start : float
        Start time in ET seconds
    window_end : float
        End time in ET seconds
    degree : int
        Degree of Chebyshev polynomials

    Returns
    -------
    tuple
        (coefficients, mid_time, half_interval)
        coefficients has shape (6, degree+1)
    """
    # Get states for this window and convert to km and km/s
    et_times = coordinates.time.et().to_numpy()
    mask = (et_times >= window_start) & (et_times <= window_end)
    states = coordinates.values[mask].copy()
    states[:, :3] *= 149597870.7  # au to km
    states[:, 3:] *= 149597870.7 / 86400.0  # au/day to km/s
    times = coordinates.time.et().to_numpy()[mask]

    # Scale time to [-1, 1] interval
    mid_time = (window_end + window_start) / 2
    half_interval = (window_end - window_start) / 2
    scaled_times = (times - mid_time) / half_interval

    # Compute Chebyshev polynomials
    T = np.zeros((len(times), degree + 1))
    T[:, 0] = 1
    T[:, 1] = scaled_times
    for i in range(2, degree + 1):
        T[:, i] = 2 * scaled_times * T[:, i - 1] - T[:, i - 2]

    # Fit polynomials to each component
    coefficients = np.zeros((6, degree + 1))
    for i in range(6):
        coefficients[i] = np.linalg.lstsq(T, states[:, i], rcond=None)[0]

    return coefficients, mid_time, half_interval


def orbits_to_spk(
    orbits: Orbits,
    output_file: str,
    start_time: Timestamp,
    end_time: Timestamp,
    propagator: Optional[Propagator] = None,
    max_processes: Optional[int] = None,
    step_days: float = 1.0 / 4,  # Every 6 hours
    target_id_start: int = 1000000,
    window_days: float = 32.0,
    comment: str = "SPK file generated from adam_core Orbits",
    kernel_type: str = "w03",
) -> Dict[str, int]:
    """
    Convert Orbits object to a SPICE SPK file using Chebyshev polynomials (Type 3).
    """
    logger.info(f"Creating SPK file: {output_file}")
    logger.info(
        f"Time range: {start_time.to_astropy().isot} to {end_time.to_astropy().isot}"
    )
    logger.info(f"Kernel type: {kernel_type}")

    # ensure SPICE is ready to go
    setup_SPICE()

    # default to a chebyshev degree of 15
    cheby_degree = 15

    # Generate propagation times
    num_steps = (
        int((end_time.mjd().to_numpy() - start_time.mjd().to_numpy()) / step_days) + 1
    )
    logger.debug(f"Generated {num_steps} time steps with step size {step_days} days")

    times = qv.concatenate(
        [start_time.add_fractional_days(i * step_days) for i in range(num_steps)]
    )
    # Propagate orbits if propagator provided
    if propagator is not None:
        logger.debug("Propagating orbits...")
        orbits = propagator.propagate_orbits(
            orbits, times, max_processes=max_processes, chunk_size=1
        )
        logger.debug("Orbit propagation complete")

    # Transform everything to a Sun origin and
    # ecliptic frame
    # Verify all orbits have the same origin
    ssb_coordinates = transform_coordinates(
        orbits.coordinates,
        frame_out="equatorial",
        origin_out=OriginCodes.SOLAR_SYSTEM_BARYCENTER,
    )

    orbits = orbits.set_column("coordinates", ssb_coordinates)

    # Create the SPK file
    handle = sp.spkopn(output_file, comment, 0)
    target_id_mappings = {}

    for i, (orbit_id, orbit) in enumerate(orbits.group_by_orbit_id()):
        logger.debug(
            f"Processing orbit {orbit_id} ({i} / ({len(orbits.orbit_id.unique())})"
        )
        # ensure orbit is sorted by time
        orbit = orbit.sort_by(["coordinates.time.days", "coordinates.time.nanos"])

        target_id = target_id_start + i

        target_id_mappings[orbit_id] = target_id
        logger.debug(f"Orbit {orbit_id} -> Target ID: {target_id}")

        # Get time range for this orbit
        orbit_start = orbit.coordinates.time.min().et()[0].as_py()
        orbit_end = orbit.coordinates.time.max().et()[0].as_py()

        if kernel_type == "w03":
            write_spkw03_segment(
                orbit,
                handle,
                target_id,
                orbit_start,
                orbit_end,
                window_days * 86400.0,
                cheby_degree,
            )
        elif kernel_type == "w09":
            write_spkw09_segment(
                orbit,
                handle,
                target_id,
                orbit_start,
                orbit_end,
            )
        else:
            raise ValueError(f"Invalid kernel type: {kernel_type}")

    # Close the SPK file
    sp.spkcls(handle)
    logger.info(f"Successfully created SPK file: {output_file}")

    return target_id_mappings


def write_spkw03_segment(
    propagated_orbit: Orbits,
    handle: int,
    target_id: int,
    start_time: float,
    end_time: float,
    window_seconds: float = 86400.0,
    cheby_degree: int = 15,
) -> None:
    logger.debug(f"Writing SPK type 03 segment for target ID {target_id}")
    # assert orbit id is unique
    assert propagated_orbit.orbit_id.unique()

    spk_frame = (
        "J2000" if propagated_orbit.coordinates.frame == "equatorial" else "ECLIPJ2000"
    )

    segment_id = (
        str(propagated_orbit.orbit_id[0].as_py())
        + "_"
        + str(start_time)
        + "_"
        + str(end_time)
    )

    # SPICE has a limit of 40 characters for the segment id
    segment_id = segment_id[:40]

    num_windows = int(np.ceil((end_time - start_time) / window_seconds))
    logger.debug(f"Fitting {num_windows} Chebyshev windows")

    # Initialize arrays for SPKW03
    cheby_coeffs = []
    window_starts = []

    for w in range(num_windows):

        start_time_window = start_time + w * window_seconds
        end_time_window = min(start_time_window + window_seconds, end_time)

        coeffs, mid_time, half_interval = fit_chebyshev(
            propagated_orbit.coordinates,
            start_time_window,
            end_time_window,
            cheby_degree,
        )

        cheby_coeffs.append(coeffs.flatten())
        window_starts.append(start_time)

    # Convert to numpy arrays
    cheby_coeffs = np.array(cheby_coeffs)
    window_starts = np.array(window_starts)

    # Write the SPKW03 segment
    sp.spkw03(
        handle,
        target_id,
        propagated_orbit.coordinates.origin.as_OriginCodes().value,
        spk_frame,
        start_time,
        end_time,
        segment_id,
        window_seconds,
        len(cheby_coeffs),
        cheby_degree,
        cheby_coeffs.flatten(),
        window_starts[0],
    )


def write_spkw09_segment(
    propagated_orbit: Orbits,
    handle: int,
    target_id: int,
    start_time: float,
    end_time: float,
) -> None:
    logger.debug(f"Writing SPK type 09 segment for target ID {target_id}")
    # assert orbit id is unique
    assert propagated_orbit.orbit_id.unique()

    spk_frame = (
        "J2000" if propagated_orbit.coordinates.frame == "equatorial" else "ECLIPJ2000"
    )

    segment_id = (
        str(propagated_orbit.orbit_id[0].as_py())
        + "_"
        + str(start_time)
        + "_"
        + str(end_time)
    )

    # SPICE has a limit of 40 characters for the segment id
    segment_id = segment_id[:40]

    epochs_tdb = (
        propagated_orbit.coordinates.time.rescale("tdb")
        .jd()
        .to_numpy(zero_copy_only=False)
    )
    epochs_et = np.array([sp.str2et(f"JD {i:.15f} TDB".format(i)) for i in epochs_tdb])

    states = propagated_orbit.coordinates.values
    states[:, 0:6] *= u.au.to(u.km)
    states[:, 3:6] /= (u.d).to(u.s)

    sp.spkw09(
        handle,
        target_id,
        propagated_orbit.coordinates.origin.as_OriginCodes().value,
        spk_frame,
        epochs_et[0],
        epochs_et[-1],
        segment_id,
        15,
        len(epochs_et),
        np.ascontiguousarray(states),
        epochs_et,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Final


@dataclass(frozen=True)
class TemplateSpec:
    template_id: str
    weight_C: float
    weight_S: float
    citation: str


# Initial shipped templates. Mixes are defined as linear combinations of the base
# C and S reflectance templates.
TEMPLATE_SPECS: Final[tuple[TemplateSpec, ...]] = (
    TemplateSpec(
        template_id="C",
        weight_C=1.0,
        weight_S=0.0,
        citation="Bus–DeMeo taxonomy (C-type) — simplified reflectance template (vendored).",
    ),
    TemplateSpec(
        template_id="S",
        weight_C=0.0,
        weight_S=1.0,
        citation="Bus–DeMeo taxonomy (S-type) — simplified reflectance template (vendored).",
    ),
    TemplateSpec(
        template_id="NEO",
        weight_C=0.5,
        weight_S=0.5,
        citation="NEO population mix (assumed): 50% C / 50% S.",
    ),
    TemplateSpec(
        template_id="MBA",
        weight_C=0.7,
        weight_S=0.3,
        citation="Main-belt population mix (assumed): 70% C / 30% S.",
    ),
)


# MPC observatory codes to treat as ATLAS in v1.
ATLAS_MPC_CODES: Final[tuple[str, ...]] = ("T08", "T05", "M22", "W68")



import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pytest
import spiceypy as sp

from ...coordinates.origin import OriginCodes
from ...time import Timestamp
from ...utils.spice import get_perturber_state, get_spice_body_state
from ..observers import OBSERVATORY_CODES, OBSERVATORY_PARALLAX_COEFFICIENTS, Observers
from ..state import get_mpc_observer_state, get_observer_state


@pytest.fixture
def codes_times() -> tuple[pa.Array, Timestamp]:
    codes = pa.array(
        ["500", "X05", "I41", "X05", "I41", "W84", "500"],
    )

    times = Timestamp.from_kwargs(
        days=[59000, 59001, 59002, 59003, 59004, 59005, 59006],
        nanos=[0, 0, 0, 0, 0, 0, 0],
        scale="tdb",
    )
    return codes, times


def test_Observers_from_codes(codes_times) -> None:
    # Test that observers from code returns the correct number of observers
    # and in the order that they were requested
    codes, times = codes_times

    observers = Observers.from_codes(codes, times)
    assert len(observers) == 7
    assert pc.all(pc.equal(observers.code, codes)).as_py()
    assert pc.all(pc.equal(observers.coordinates.time.days, times.days)).as_py()
    assert pc.all(pc.equal(observers.coordinates.time.nanos, times.nanos)).as_py()


def test_Observers_from_codes_non_pyarrow(codes_times) -> None:
    # Test that observers from code returns the correct number of observers
    # and in the order that they were requested
    codes, times = codes_times

    observers = Observers.from_codes(codes.to_numpy(zero_copy_only=False), times)
    assert len(observers) == 7
    assert pc.all(pc.equal(observers.code, codes)).as_py()
    assert pc.all(pc.equal(observers.coordinates.time.days, times.days)).as_py()
    assert pc.all(pc.equal(observers.coordinates.time.nanos, times.nanos)).as_py()

    observers = Observers.from_codes(codes.to_pylist(), times)
    assert len(observers) == 7
    assert pc.all(pc.equal(observers.code, codes)).as_py()
    assert pc.all(pc.equal(observers.coordinates.time.days, times.days)).as_py()
    assert pc.all(pc.equal(observers.coordinates.time.nanos, times.nanos)).as_py()


def test_Observers_from_codes_raises(codes_times) -> None:
    # Test that observers from code raises an error if the codes and times
    # are not the same length
    codes, times = codes_times

    with pytest.raises(ValueError, match="codes and times must have the same length."):
        Observers.from_codes(codes[:3], times)
    with pytest.raises(ValueError, match="codes and times must have the same length."):
        Observers.from_codes(codes, times[:3])


def test_ObservatoryParallaxCoefficients_lon_lat() -> None:
    # Data taken from: https://en.wikipedia.org/wiki/List_of_observatory_codes
    # From: https://geohack.toolforge.org/geohack.php?pagename=Zwicky_Transient_Facility&params=33.35731_N_116.85981_W_
    I41 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "I41")
    lon, lat = I41.lon_lat()
    np.testing.assert_allclose(lon[0], -116.85981, atol=1e-4, rtol=0)
    np.testing.assert_allclose(lat[0], 33.35731, atol=1e-4, rtol=0)

    # From: https://geohack.toolforge.org/geohack.php?pagename=Vera_C._Rubin_Observatory&params=30_14_40.7_S_70_44_57.9_W_region:CL-CO_type:landmark
    X05 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "X05")
    lon, lat = X05.lon_lat()
    np.testing.assert_allclose(lon[0], -70.749417, atol=1e-4, rtol=0)
    np.testing.assert_allclose(lat[0], -30.244639, atol=1e-4, rtol=0)

    # From: https://geohack.toolforge.org/geohack.php?pagename=V%C3%ADctor_M._Blanco_Telescope&params=30.16967_S_70.80653_W_
    W84 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "W84")
    lon, lat = W84.lon_lat()
    np.testing.assert_allclose(lon[0], -70.80653, atol=1e-3, rtol=0)
    np.testing.assert_allclose(lat[0], -30.169679, atol=1e-3, rtol=0)

    # From: https://geohack.toolforge.org/geohack.php?params=32_22_48.6_S_20_48_38.1_E
    M22 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "M22")
    lon, lat = M22.lon_lat()
    np.testing.assert_allclose(lon[0], 20.810583, atol=1e-4, rtol=0)
    np.testing.assert_allclose(lat[0], -32.380167, atol=1e-4, rtol=0)

    # From: https://geohack.toolforge.org/geohack.php?params=20_42_26.04_N_156_15_21.28_W
    F51 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "F51")
    lon, lat = F51.lon_lat()
    np.testing.assert_allclose(lon[0], -156.255911, atol=1e-4, rtol=0)
    np.testing.assert_allclose(lat[0], 20.707233, atol=1e-4, rtol=0)


def test_ObservatoryParallaxCoeffiecients_timezone() -> None:
    I41 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "I41")
    assert I41.timezone() == "America/Los_Angeles"

    X05 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "X05")
    assert X05.timezone() == "America/Santiago"

    W84 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "W84")
    assert W84.timezone() == "America/Santiago"

    M22 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "M22")
    assert M22.timezone() == "Africa/Johannesburg"

    F51 = OBSERVATORY_PARALLAX_COEFFICIENTS.select("code", "F51")
    assert F51.timezone() == "Pacific/Honolulu"


def test_origincode_observer():
    """Test getting observer state using an OriginCodes enum."""
    test_time = Timestamp.from_iso8601(["2022-01-01T00:00:00Z"])

    # Test with EARTH OriginCode
    earth_code = OriginCodes.EARTH

    # Get state using get_observer_state
    observer_state = get_observer_state(
        earth_code, test_time, frame="ecliptic", origin=OriginCodes.SUN
    )

    # Get state using get_perturber_state
    perturber_state = get_perturber_state(
        earth_code, test_time, frame="ecliptic", origin=OriginCodes.SUN
    )

    # Both should return identical results
    np.testing.assert_allclose(
        observer_state.values, perturber_state.values, rtol=1e-10
    )


def test_mpc_observatory_code():
    """Test getting observer state using an MPC observatory code."""
    test_time = Timestamp.from_iso8601(["2022-01-01T00:00:00Z"])

    # Use "500" (geocenter) which should return Earth's state directly
    geocenter_state = get_observer_state(
        "500", test_time, frame="ecliptic", origin=OriginCodes.SUN
    )

    # Get Earth's state
    earth_state = get_perturber_state(
        OriginCodes.EARTH, test_time, frame="ecliptic", origin=OriginCodes.SUN
    )

    # Both should return identical results
    np.testing.assert_allclose(geocenter_state.values, earth_state.values, rtol=1e-10)

    # Test a non-geocenter MPC code if available
    non_geocenter_codes = [
        code for code in OBSERVATORY_CODES if code != "500" and len(code) == 3
    ]

    if non_geocenter_codes:
        # Use the first available MPC code
        test_code = non_geocenter_codes[0]

        # Get state using the MPC code
        mpc_state = get_observer_state(
            test_code, test_time, frame="ecliptic", origin=OriginCodes.SUN
        )

        # Verify we got valid state vectors
        assert mpc_state.values.shape == (1, 6)
        assert mpc_state.frame == "ecliptic"
        assert mpc_state.origin.code[0].as_py() == "SUN"

        # The MPC observatory state should be near Earth but not identical
        # (it's on Earth's surface)
        earth_distance = np.linalg.norm(
            mpc_state.values[0, :3] - earth_state.values[0, :3]
        )

        # Distance should be small (Earth radius is about 4.2e-5 AU)
        assert 1e-6 < earth_distance < 1e-4


def test_invalid_code():
    """Test that using an invalid code raises the correct error."""
    test_time = Timestamp.from_iso8601(["2022-01-01T00:00:00Z"])

    # Try with a completely invalid code
    with pytest.raises(ValueError, match="not a valid MPC observatory code"):
        get_observer_state(
            "INVALID_CODE", test_time, frame="ecliptic", origin=OriginCodes.SUN
        )

    # Try with an invalid type
    with pytest.raises(AssertionError, match="code must be a string or OriginCodes"):
        get_observer_state(
            123,  # Not a string or OriginCodes
            test_time,
            frame="ecliptic",
            origin=OriginCodes.SUN,
        )


def test_mpc_vs_spice_precedence():
    """Test that SPICE codes take precedence over MPC codes when both exist."""
    test_time = Timestamp.from_iso8601(["2022-01-01T00:00:00Z"])

    # For this test, we'll create a temporary SPICE kernel with a body name
    # that matches an MPC observatory code

    # First, find an MPC code that we can use
    test_mpc_code = None

    for code in OBSERVATORY_CODES:
        # Skip geocenter and spacecraft codes
        if code == "500" or len(code) != 3:
            continue

        # Check if this code isn't already defined in SPICE
        try:
            sp.bodn2c(code)
        except sp.SpiceyError:
            test_mpc_code = code
            break

    if test_mpc_code is None:
        pytest.skip("Could not find a suitable MPC code for this test")

    # Get the MPC observatory state first for comparison
    mpc_state = get_mpc_observer_state(
        test_mpc_code, test_time, frame="ecliptic", origin=OriginCodes.SUN
    )

    # Now, get Mars state with direct SPICE call
    mars_state = get_spice_body_state(
        OriginCodes.MARS_BARYCENTER.value,
        test_time,
        frame="ecliptic",
        origin=OriginCodes.SUN,
    )

    # Create a temporary alias in SPICE that maps the MPC code to Mars ID
    # This simulates having a SPICE object with the same name as an MPC code
    sp.boddef(test_mpc_code, OriginCodes.MARS_BARYCENTER.value)

    try:
        # Now get the state again, it should still return MPC state
        # as MPC codes take precedence over SPICE codes
        hybrid_state = get_observer_state(
            test_mpc_code, test_time, frame="ecliptic", origin=OriginCodes.SUN
        )

        # The result should match Mars, not the MPC observatory
        np.testing.assert_allclose(hybrid_state.values, mpc_state.values, rtol=1e-10)

        # And it should NOT match the MPC observatory
        with pytest.raises(AssertionError):
            np.testing.assert_allclose(
                hybrid_state.values, mars_state.values, rtol=1e-10
            )
    finally:
        # Clean up the temporary alias to avoid affecting other tests
        try:
            # Try to undefine the alias
            sp.bodc2n(
                OriginCodes.MARS_BARYCENTER.value
            )  # This will raise an error, but we need to try something
        except sp.SpiceyError:
            pass  # Ignore, we just want to make sure we don't leave aliases



import numpy as np
import pyarrow as pa
import pytest
from adam_assist import ASSISTPropagator

from adam_core.propagator import propagator as propagator_module
from adam_core.ray_cluster import initialize_use_ray

from ...observers.observers import Observers
from ...time import Timestamp
from ...utils.helpers.orbits import make_real_orbits


@pytest.mark.benchmark(group="assist_ephemeris")
def test_assist_generate_ephemeris_single_process_benchmark(benchmark):
    """
    Benchmark the CPU-bound ephemeris generation path implemented by
    ASSISTPropagator using a single process.
    """
    # Keep sizes modest to stay CI-friendly while still exercising the code path
    num_orbits = 20
    num_times = 20

    orbits = make_real_orbits(num_orbits)
    # Anchor benchmark times near the input orbit epochs. Some sample orbits have epochs
    # far from an arbitrary fixed time (e.g., MJD=60000). Long propagations can produce
    # physically unrealistic states (e.g., close encounters/collisions in N-body integration)
    # which then break light-time correction (NaN/inf). Using a per-sample epoch keeps the
    # benchmark stable across integrator/library versions while exercising the same code path.
    orbit_mjd = orbits.coordinates.time.mjd().to_numpy(zero_copy_only=False)
    base_mjd = int(round(float(np.median(orbit_mjd))))
    times = Timestamp.from_mjd(np.arange(base_mjd, base_mjd + num_times), scale="tdb")
    observers = Observers.from_code("500", times)

    propagator = ASSISTPropagator()

    def run():
        return propagator.generate_ephemeris(
            orbits,
            observers,
            covariance=True,
            num_samples=3,
            max_processes=1,
            chunk_size=1,
            seed=42,
        )

    ephemeris = benchmark(run)
    assert len(ephemeris) == len(orbits) * len(observers)


@pytest.mark.benchmark(group="assist_ephemeris")
def test_assist_generate_ephemeris_multi_process_benchmark(benchmark):
    """
    Benchmark the ephemeris generation with multi-processing enabled (Ray).
    """
    num_orbits = 20
    num_times = 20

    orbits = make_real_orbits(num_orbits)
    orbit_mjd = orbits.coordinates.time.mjd().to_numpy(zero_copy_only=False)
    base_mjd = int(round(float(np.median(orbit_mjd))))
    times = Timestamp.from_mjd(np.arange(base_mjd, base_mjd + num_times), scale="tdb")
    observers = Observers.from_code("500", times)

    propagator = ASSISTPropagator()
    initialize_use_ray(num_cpus=4)

    def run():
        return propagator.generate_ephemeris(
            orbits,
            observers,
            covariance=True,
            num_samples=3,
            max_processes=4,
            chunk_size=1,
            seed=42,
        )

    ephemeris = benchmark(run)
    assert len(ephemeris) == len(orbits) * len(observers)


@pytest.mark.benchmark(group="observer_alignment")
@pytest.mark.parametrize("n", [10_000])
def test_observer_alignment_indices_benchmark(benchmark, n: int):
    """
    Benchmark observer-row alignment used for predicted magnitude attachment.

    We intentionally benchmark the alignment indices computation in isolation so we can
    swap strategies without involving propagation or coordinate transforms.
    """
    rng = np.random.default_rng(123)
    codes = np.asarray([f"C{i:02d}" for i in range(25)], dtype=object)
    obs_code = rng.choice(codes, size=n, replace=True)
    # Make keys unique by construction using unique days.
    obs_days = (60000 + np.arange(n)).astype(np.int64)
    obs_nanos = np.zeros(n, dtype=np.int64)

    perm = rng.permutation(n)
    eph_code = obs_code[perm]
    eph_days = obs_days[perm]
    eph_nanos = obs_nanos[perm]

    def run_struct():
        idx = propagator_module._alignment_indices_struct_index_in(
            obs_code=pa.array(obs_code),
            obs_days=pa.array(obs_days),
            obs_nanos=pa.array(obs_nanos),
            eph_code=pa.array(eph_code),
            eph_days=pa.array(eph_days),
            eph_nanos=pa.array(eph_nanos),
        )
        return idx

    def run_string():
        idx = propagator_module._alignment_indices_string_keys(
            obs_code=obs_code,
            obs_days=obs_days,
            obs_nanos=obs_nanos,
            eph_code=eph_code,
            eph_days=eph_days,
            eph_nanos=eph_nanos,
        )
        return idx

    idx_struct = benchmark(run_struct)
    # Validate correctness once (outside the timed portion) and sanity-check string impl too.
    np.testing.assert_array_equal(obs_days[idx_struct], eph_days)
    idx_string = run_string()
    np.testing.assert_array_equal(obs_days[idx_string], eph_days)


@pytest.mark.benchmark(group="hg_param_mapping")
@pytest.mark.parametrize("n_unique,n_rep", [(20_000, 5)])
def test_hg_param_mapping_benchmark(benchmark, n_unique: int, n_rep: int):
    """
    Benchmark orbit_id -> (H_v, G) mapping aligned to ephemeris rows.

    This mirrors the common case where propagated orbits/ephemeris have repeated orbit_ids
    across many epochs, but physical parameters are constant per orbit_id.
    """
    rng = np.random.default_rng(321)
    unique_ids = np.asarray([f"o{i:06d}" for i in range(n_unique)], dtype=object)
    # Repeated rows as typically produced by propagation to many times.
    orbit_id = np.repeat(unique_ids, n_rep)
    H_v = np.repeat(rng.uniform(10.0, 25.0, size=n_unique).astype(np.float64), n_rep)
    G = np.repeat(rng.uniform(0.0, 0.5, size=n_unique).astype(np.float64), n_rep)

    # Ephemeris orbit_id is a shuffled view of the repeated orbit_id.
    eph_orbit_id = orbit_id[rng.permutation(len(orbit_id))]

    orbit_id_a = pa.array(orbit_id)
    H_v_a = pa.array(H_v)
    G_a = pa.array(G)
    eph_orbit_id_a = pa.array(eph_orbit_id)

    # Smoke-check: output arrays match expected length and have finite values where expected.
    H_ref, G_ref = propagator_module._hg_params_for_ephemeris_rows_arrow(
        orbit_id=orbit_id_a,
        H_v=H_v_a,
        G=G_a,
        ephemeris_orbit_id=eph_orbit_id_a,
    )
    assert len(H_ref) == len(eph_orbit_id)
    assert len(G_ref) == len(eph_orbit_id)
    assert np.all(np.isfinite(H_ref))
    assert np.all(np.isfinite(G_ref))

    def run():
        return propagator_module._hg_params_for_ephemeris_rows_arrow(
            orbit_id=orbit_id_a,
            H_v=H_v_a,
            G=G_a,
            ephemeris_orbit_id=eph_orbit_id_a,
        )

    H_out, G_out = benchmark(run)
    np.testing.assert_allclose(H_out, H_ref, equal_nan=True)
    np.testing.assert_allclose(G_out, G_ref, equal_nan=True)



import quivr as qv


class PhysicalParameters(qv.Table):
    """
    Physical parameters required for photometric magnitude predictions.

    Notes
    -----
    Currently, absolute magnitude is assumed to be in Johnson-Cousins V-band.
    """

    # Absolute magnitude (V-band)
    H_v = qv.Float64Column(nullable=True)

    # 1-sigma uncertainty in H_v (mag)
    H_v_sigma = qv.Float64Column(nullable=True)

    # Photometric slope parameter (H-G system)
    G = qv.Float64Column(nullable=True)

    # 1-sigma uncertainty in G
    G_sigma = qv.Float64Column(nullable=True)

    # Optional fit diagnostics (useful when these parameters are estimated from data)
    #
    # - `sigma_eff` is an empirical residual scatter estimate (mag). If per-point `mag_sigma`
    #   is missing, this will also absorb model error (e.g., incorrect fixed G).
    sigma_eff = qv.Float64Column(nullable=True)
    chi2_red = qv.Float64Column(nullable=True)



"""
Tools for building (vendoring) bandpass curve and reflectance template data.

Bandpass curves are sourced from the SVO Filter Profile Service. Please see
`REFERENCES.md` for the required acknowledgement and citations when using this
service.
"""

# coverage: ignore-file

from __future__ import annotations

import io
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import pyarrow.parquet as pq
import requests
from astropy.io import fits
from astropy.io.votable import parse_single_table

from .tables import (
    AsteroidTemplates,
    BandpassCurves,
    ObservatoryBandMap,
    TemplateBandpassIntegrals,
)


@dataclass(frozen=True)
class _SvoBandpassSpec:
    filter_id: str
    instrument: str
    band: str
    svo_id: str


def _default_svo_bandpass_specs() -> tuple[_SvoBandpassSpec, ...]:
    # NOTE: These SVO IDs were validated via fps.php?ID=<id> returning VOTable with
    # columns Wavelength (Angstrom) and Transmission.
    return (
        # Standard Johnson/Cousins-like passbands (Bessell). Note: `V` is kept for
        # backwards compatibility; it is sourced from Bessell.V.
        _SvoBandpassSpec("V", "Bessell", "V", "Generic/Bessell.V"),
        _SvoBandpassSpec("Bessell_U", "Bessell", "U", "Generic/Bessell.U"),
        _SvoBandpassSpec("Bessell_B", "Bessell", "B", "Generic/Bessell.B"),
        _SvoBandpassSpec("Bessell_R", "Bessell", "R", "Generic/Bessell.R"),
        _SvoBandpassSpec("Bessell_I", "Bessell", "I", "Generic/Bessell.I"),
        # Rubin / LSST
        _SvoBandpassSpec("LSST_u", "LSST", "u", "LSST/LSST.u"),
        _SvoBandpassSpec("LSST_g", "LSST", "g", "LSST/LSST.g"),
        _SvoBandpassSpec("LSST_r", "LSST", "r", "LSST/LSST.r"),
        _SvoBandpassSpec("LSST_i", "LSST", "i", "LSST/LSST.i"),
        _SvoBandpassSpec("LSST_z", "LSST", "z", "LSST/LSST.z"),
        _SvoBandpassSpec("LSST_y", "LSST", "y", "LSST/LSST.y"),
        # Sloan Digital Sky Survey (SDSS) ugriz
        _SvoBandpassSpec("SDSS_u", "SDSS", "u", "SLOAN/SDSS.u"),
        _SvoBandpassSpec("SDSS_g", "SDSS", "g", "SLOAN/SDSS.g"),
        _SvoBandpassSpec("SDSS_r", "SDSS", "r", "SLOAN/SDSS.r"),
        _SvoBandpassSpec("SDSS_i", "SDSS", "i", "SLOAN/SDSS.i"),
        _SvoBandpassSpec("SDSS_z", "SDSS", "z", "SLOAN/SDSS.z"),
        # Pan-STARRS1 (PS1) grizy
        _SvoBandpassSpec("PS1_g", "PS1", "g", "PAN-STARRS/PS1.g"),
        _SvoBandpassSpec("PS1_r", "PS1", "r", "PAN-STARRS/PS1.r"),
        _SvoBandpassSpec("PS1_i", "PS1", "i", "PAN-STARRS/PS1.i"),
        _SvoBandpassSpec("PS1_z", "PS1", "z", "PAN-STARRS/PS1.z"),
        _SvoBandpassSpec("PS1_y", "PS1", "y", "PAN-STARRS/PS1.y"),
        # ZTF
        _SvoBandpassSpec("ZTF_g", "ZTF", "g", "Palomar/ZTF.g"),
        _SvoBandpassSpec("ZTF_r", "ZTF", "r", "Palomar/ZTF.r"),
        _SvoBandpassSpec("ZTF_i", "ZTF", "i", "Palomar/ZTF.i"),
        # DECam
        _SvoBandpassSpec("DECam_u", "DECam", "u", "CTIO/DECam.u"),
        _SvoBandpassSpec("DECam_g", "DECam", "g", "CTIO/DECam.g"),
        _SvoBandpassSpec("DECam_r", "DECam", "r", "CTIO/DECam.r"),
        _SvoBandpassSpec("DECam_i", "DECam", "i", "CTIO/DECam.i"),
        _SvoBandpassSpec("DECam_z", "DECam", "z", "CTIO/DECam.z"),
        _SvoBandpassSpec("DECam_Y", "DECam", "Y", "CTIO/DECam.Y"),
        _SvoBandpassSpec("DECam_VR", "DECam", "VR", "CTIO/DECam.VR_filter"),
        # Mosaic3 (Mayall / KPNO): SVO provides the Mosaic3 z-band passband via MzLS.
        _SvoBandpassSpec("Mosaic3_z", "Mosaic3", "z", "KPNO/MzLS.z"),
        # Bok / 90Prime (BASS) g/r used by MPC code V00 (Kitt Peak-Bok).
        _SvoBandpassSpec("BASS_g", "BASS", "g", "BOK/BASS.g"),
        _SvoBandpassSpec("BASS_r", "BASS", "r", "BOK/BASS.r"),
        # SkyMapper
        _SvoBandpassSpec("SkyMapper_u", "SkyMapper", "u", "SkyMapper/SkyMapper.u"),
        _SvoBandpassSpec("SkyMapper_v", "SkyMapper", "v", "SkyMapper/SkyMapper.v"),
        _SvoBandpassSpec("SkyMapper_g", "SkyMapper", "g", "SkyMapper/SkyMapper.g"),
        _SvoBandpassSpec("SkyMapper_r", "SkyMapper", "r", "SkyMapper/SkyMapper.r"),
        _SvoBandpassSpec("SkyMapper_i", "SkyMapper", "i", "SkyMapper/SkyMapper.i"),
        _SvoBandpassSpec("SkyMapper_z", "SkyMapper", "z", "SkyMapper/SkyMapper.z"),
        # ATLAS
        # SVO uses the names "Atlas.cyan" and "Atlas.orange" for c/o.
        _SvoBandpassSpec("ATLAS_c", "ATLAS", "c", "Misc/Atlas.cyan"),
        _SvoBandpassSpec("ATLAS_o", "ATLAS", "o", "Misc/Atlas.orange"),
    )


def _download_svo_votable(svo_id: str, *, timeout_s: int = 60) -> bytes:
    url = f"https://svo2.cab.inta-csic.es/theory/fps/fps.php?ID={svo_id}"
    resp = requests.get(url, timeout=timeout_s)
    resp.raise_for_status()
    return resp.content


def _parse_svo_curve(
    votable_bytes: bytes,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    table = parse_single_table(io.BytesIO(votable_bytes)).to_table()
    if "Wavelength" not in table.colnames or "Transmission" not in table.colnames:
        raise ValueError(f"Unexpected SVO schema; columns={table.colnames}")

    wl = np.asarray(table["Wavelength"], dtype=np.float64)
    trans = np.asarray(table["Transmission"], dtype=np.float64)
    if wl.ndim != 1 or trans.ndim != 1:
        raise ValueError("Unexpected SVO curve dimensions")
    if len(wl) != len(trans):
        raise ValueError("SVO Wavelength and Transmission lengths do not match")
    if len(wl) < 2:
        raise ValueError("SVO curve has too few points")

    # SVO FPS provides wavelength in Angstrom for these filters.
    wl_nm = wl / 10.0
    return wl_nm, trans


def _normalize_curve(
    wavelength_nm: npt.NDArray[np.float64], throughput: npt.NDArray[np.float64]
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    wl = np.asarray(wavelength_nm, dtype=np.float64)
    thr = np.asarray(throughput, dtype=np.float64)
    if np.any(~np.isfinite(wl)) or np.any(~np.isfinite(thr)):
        raise ValueError("Non-finite values in curve")

    order = np.argsort(wl)
    wl = wl[order]
    thr = thr[order]

    # Drop duplicate wavelengths by keeping the first occurrence.
    unique_mask = np.concatenate(([True], np.diff(wl) > 0))
    wl = wl[unique_mask]
    thr = thr[unique_mask]
    if len(wl) < 2:
        raise ValueError("Curve is degenerate after de-duplication")

    thr = np.clip(thr, 0.0, 1.0)
    m = float(np.max(thr))
    if m <= 0.0:
        raise ValueError("Curve has zero throughput everywhere")
    thr = thr / m
    return wl, thr


def build_bandpass_curves(
    out_dir: Path,
    *,
    specs: Iterable[_SvoBandpassSpec] | None = None,
    timeout_s: int = 60,
) -> BandpassCurves:
    """
    Download, normalize, and write `BandpassCurves` to `out_dir/bandpass_curves.parquet`.
    """
    if specs is None:
        specs = _default_svo_bandpass_specs()

    records = list(specs)
    if not records:
        raise ValueError("No bandpass specs provided")

    filter_ids: list[str] = []
    instruments: list[str] = []
    bands: list[str] = []
    sources: list[str] = []
    wavelengths: list[npt.NDArray[np.float64]] = []
    throughputs: list[npt.NDArray[np.float64]] = []

    seen: set[str] = set()
    for spec in records:
        if spec.filter_id in seen:
            raise ValueError(f"Duplicate filter_id in specs: {spec.filter_id}")
        seen.add(spec.filter_id)

        raw = _download_svo_votable(spec.svo_id, timeout_s=timeout_s)
        wl_nm, thr = _parse_svo_curve(raw)
        wl_nm, thr = _normalize_curve(wl_nm, thr)

        filter_ids.append(spec.filter_id)
        instruments.append(spec.instrument)
        bands.append(spec.band)
        sources.append(
            f"https://svo2.cab.inta-csic.es/theory/fps/fps.php?ID={spec.svo_id}"
        )
        wavelengths.append(wl_nm)
        throughputs.append(thr)

    wl_arr = pa.array(wavelengths, type=pa.large_list(pa.float64()))
    thr_arr = pa.array(throughputs, type=pa.large_list(pa.float64()))

    table = BandpassCurves.from_kwargs(
        filter_id=filter_ids,
        instrument=instruments,
        band=bands,
        wavelength_nm=wl_arr,
        throughput=thr_arr,
        source=sources,
    )

    out_path = out_dir.joinpath("bandpass_curves.parquet")
    table.to_parquet(out_path)
    return table


def build_observatory_band_map(out_dir: Path) -> ObservatoryBandMap:
    """
    Write `ObservatoryBandMap` to `out_dir/observatory_band_map.parquet`.
    """
    mappings: list[tuple[str, str, str]] = []

    # DECam (W84)
    for band in ["u", "g", "r", "i", "z"]:
        mappings.append(("W84", band, f"DECam_{band}"))
    mappings.append(("W84", "Y", "DECam_Y"))
    mappings.append(("W84", "y", "DECam_Y"))  # pragmatic alias
    mappings.append(("W84", "VR", "DECam_VR"))
    mappings.append(("W84", "vr", "DECam_VR"))  # pragmatic alias

    # Mosaic3 (695): z-band only in SVO for Mosaic3/MzLS.
    mappings.append(("695", "z", "Mosaic3_z"))

    # ZTF (I41)
    for band in ["g", "r", "i"]:
        mappings.append(("I41", band, f"ZTF_{band}"))

    # SkyMapper (Q55)
    mappings.extend(
        [
            ("Q55", "u", "SkyMapper_u"),
            ("Q55", "v", "SkyMapper_v"),
            ("Q55", "g", "SkyMapper_g"),
            ("Q55", "r", "SkyMapper_r"),
            ("Q55", "i", "SkyMapper_i"),
            ("Q55", "z", "SkyMapper_z"),
        ]
    )

    # Rubin/LSST (X05)
    for band in ["u", "g", "r", "i", "z", "y"]:
        mappings.append(("X05", band, f"LSST_{band}"))
    mappings.append(("X05", "Y", "LSST_y"))  # pragmatic alias

    # ATLAS (multiple MPC codes share the same c/o passbands)
    for code in ["T08", "T05", "M22", "W68"]:
        mappings.append((code, "c", "ATLAS_c"))
        mappings.append((code, "o", "ATLAS_o"))

    # Kitt Peak-Bok (V00): BASS / Bok-90Prime g/r.
    mappings.append(("V00", "g", "BASS_g"))
    mappings.append(("V00", "r", "BASS_r"))

    obs_codes, bands, filter_ids = zip(*mappings)
    keys = [f"{c}|{b}" for c, b in zip(obs_codes, bands)]
    table = ObservatoryBandMap.from_kwargs(
        observatory_code=list(obs_codes),
        reported_band=list(bands),
        filter_id=list(filter_ids),
        key=list(keys),
    )
    out_path = out_dir.joinpath("observatory_band_map.parquet")
    table.to_parquet(out_path)
    return table


def build_asteroid_templates(out_dir: Path) -> AsteroidTemplates:
    """
    Build and vendor asteroid reflectance templates (C, S) plus fixed mixes (NEO, MBA).

    Notes
    -----
    - These are intentionally simple, hand-built reflectance templates designed for
      early synthetic photometry and are *not* a replacement for higher-fidelity
      published average spectra.
    - Reflectance is normalized to 1.0 at 550 nm.
    """
    # Shared wavelength grid (nm) spanning typical optical filters.
    wl = np.linspace(300.0, 1100.0, 4000, dtype=np.float64)

    def _normalize_at_550(
        reflectance: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        r550 = float(np.interp(550.0, wl, reflectance))
        if r550 <= 0.0 or not np.isfinite(r550):
            raise ValueError("Template reflectance is invalid at 550 nm")
        return (reflectance / r550).astype(np.float64, copy=False)

    # C-type: relatively flat/slightly red with a weak 0.7 µm feature and UV drop-off.
    c = np.ones_like(wl)
    c *= 1.0 + 0.15 * (wl - 550.0) / 550.0
    c *= 1.0 - 0.03 * np.exp(-0.5 * ((wl - 700.0) / 50.0) ** 2)
    uv = wl < 450.0
    c[uv] *= np.exp(-(((450.0 - wl[uv]) / 100.0) ** 2))
    c = _normalize_at_550(c)

    # S-type: redder slope with onset of the ~1 µm silicate absorption and UV drop-off.
    s = np.ones_like(wl)
    s *= 1.0 + 0.50 * (wl - 550.0) / 550.0
    s *= 1.0 - 0.10 * np.exp(-0.5 * ((wl - 950.0) / 100.0) ** 2)
    uv = wl < 500.0
    s[uv] *= np.exp(-(((500.0 - wl[uv]) / 120.0) ** 2))
    s = _normalize_at_550(s)

    # Mixes are linear combinations in reflectance space.
    neo = 0.5 * c + 0.5 * s
    mba = 0.7 * c + 0.3 * s

    templates: list[tuple[str, float, float, npt.NDArray[np.float64]]] = [
        ("C", 1.0, 0.0, c),
        ("S", 0.0, 1.0, s),
        ("NEO", 0.5, 0.5, neo),
        ("MBA", 0.7, 0.3, mba),
    ]

    citations: dict[str, str] = {
        "C": "Hand-built C-type reflectance template (optical/NIR). Informed by mean asteroid colors (e.g., Bowell & Lumme 1979; Erasmus et al. 2019). Normalized at 550 nm.",
        "S": "Hand-built S-type reflectance template (optical/NIR). Informed by mean asteroid colors (e.g., Bowell & Lumme 1979; Erasmus et al. 2019). Normalized at 550 nm.",
        "NEO": "Assumed NEO population mix: 50% C / 50% S (linear reflectance mix).",
        "MBA": "Assumed main-belt population mix: 70% C / 30% S (linear reflectance mix).",
    }

    wl_arr = pa.array([wl] * len(templates), type=pa.large_list(pa.float64()))
    refl_arr = pa.array([t[3] for t in templates], type=pa.large_list(pa.float64()))
    table = AsteroidTemplates.from_kwargs(
        template_id=[t[0] for t in templates],
        wavelength_nm=wl_arr,
        reflectance=refl_arr,
        weight_C=[t[1] for t in templates],
        weight_S=[t[2] for t in templates],
        citation=[citations[t[0]] for t in templates],
    )

    out_path = out_dir.joinpath("asteroid_templates.parquet")
    table.to_parquet(out_path)
    return table


def build_solar_spectrum(out_dir: Path) -> pa.Table:
    """
    Download and vendor a fixed solar spectrum used for synthetic photometry integrals.

    Source
    ------
    STScI reference atlas solar spectrum FITS used widely in calibration contexts.
    """
    url = "https://archive.stsci.edu/hlsps/reference-atlases/cdbs/grid/solsys/solar_spec.fits"
    resp = requests.get(url, timeout=120)
    resp.raise_for_status()

    with fits.open(io.BytesIO(resp.content)) as hdul:
        data = hdul[1].data
        wavelength_angstrom = np.asarray(data["WAVELENGTH"], dtype=np.float64)
        flux = np.asarray(data["FLUX"], dtype=np.float64)

    wavelength_nm = wavelength_angstrom / 10.0

    # Restrict to a useful range for optical/NIR passbands, then normalize.
    mask = (wavelength_nm >= 300.0) & (wavelength_nm <= 1200.0)
    wavelength_nm = wavelength_nm[mask]
    flux = flux[mask]
    flux = flux / float(np.max(flux))

    table = pa.table(
        {
            "wavelength_nm": pa.array(wavelength_nm, type=pa.float64()),
            "flux": pa.array(flux, type=pa.float64()),
        }
    )
    out_path = out_dir.joinpath("solar_spectrum.parquet")
    pq.write_table(table, out_path)
    return table


def build_template_bandpass_integrals(out_dir: Path) -> TemplateBandpassIntegrals:
    """
    Compute and vendor template×filter integrals for photon-counting synthetic photometry.
    """
    curves = BandpassCurves.from_parquet(out_dir.joinpath("bandpass_curves.parquet"))
    templates = AsteroidTemplates.from_parquet(
        out_dir.joinpath("asteroid_templates.parquet")
    )
    solar_tbl = pq.read_table(out_dir.joinpath("solar_spectrum.parquet"))
    solar_wl = np.asarray(
        solar_tbl["wavelength_nm"].to_numpy(zero_copy_only=False), dtype=np.float64
    )
    solar_flux = np.asarray(
        solar_tbl["flux"].to_numpy(zero_copy_only=False), dtype=np.float64
    )

    if len(solar_wl) < 2 or np.any(np.diff(solar_wl) <= 0):
        order = np.argsort(solar_wl)
        solar_wl = solar_wl[order]
        solar_flux = solar_flux[order]

    def _interp(
        x: npt.NDArray[np.float64],
        y: npt.NDArray[np.float64],
        x_new: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        return np.interp(x_new, x, y, left=0.0, right=0.0).astype(
            np.float64, copy=False
        )

    template_ids: list[str] = []
    filter_ids: list[str] = []
    integrals: list[float] = []

    for tmpl_id, wl_list, refl_list in zip(
        templates.template_id.to_pylist(),
        templates.wavelength_nm.to_pylist(),
        templates.reflectance.to_pylist(),
    ):
        tmpl_wl = np.asarray(wl_list, dtype=np.float64)
        tmpl_refl = np.asarray(refl_list, dtype=np.float64)
        for filt_id, filt_wl_list, thr_list in zip(
            curves.filter_id.to_pylist(),
            curves.wavelength_nm.to_pylist(),
            curves.throughput.to_pylist(),
        ):
            filt_wl = np.asarray(filt_wl_list, dtype=np.float64)
            filt_thr = np.asarray(thr_list, dtype=np.float64)

            wl_min = max(
                float(solar_wl.min()), float(tmpl_wl.min()), float(filt_wl.min())
            )
            wl_max = min(
                float(solar_wl.max()), float(tmpl_wl.max()), float(filt_wl.max())
            )
            if wl_max <= wl_min:
                val = float("nan")
            else:
                mask = (solar_wl >= wl_min) & (solar_wl <= wl_max)
                wl = solar_wl[mask]
                sun = solar_flux[mask]
                t = _interp(filt_wl, filt_thr, wl)
                r = _interp(tmpl_wl, tmpl_refl, wl)
                val = float(np.trapezoid(sun * r * t * wl, wl))

            template_ids.append(str(tmpl_id))
            filter_ids.append(str(filt_id))
            integrals.append(val)

    table = TemplateBandpassIntegrals.from_kwargs(
        template_id=template_ids,
        filter_id=filter_ids,
        integral_photon=integrals,
    )
    out_path = out_dir.joinpath("template_bandpass_integrals.parquet")
    table.to_parquet(out_path)
    return table



"""
Utilities for generating and reading OpenSpace Asset files
"""

import os
from dataclasses import dataclass
from typing import List, Literal, Optional, Tuple, Union

import numpy as np
import pyarrow as pa

from ...constants import KM_P_AU, S_P_DAY
from ...coordinates.keplerian import KeplerianCoordinates
from ...coordinates.origin import OriginCodes
from ...coordinates.transform import transform_coordinates
from ...orbits import Orbits
from .lua import LuaDict
from .renderable import (
    Renderable,
    RenderableOrbitalKepler,
    RenderableOrbitalKeplerFormat,
    RenderableOrbitalKeplerRendering,
    RenderableTrailOrbit,
    RenderableTrailRendering,
    RenderBinMode,
    Resource,
)
from .translation import KeplerTranslation, SpiceTranslation, Transform


def _safe_orbital_period(period: float) -> float:
    """
    For an orbit with no period or infinite period, use a default value in days.
    """
    orbital_period = float(period)
    # Use a default value if orbital period is infinite or NaN
    if np.isnan(orbital_period) or np.isinf(orbital_period):
        return 10000.0
    return orbital_period


@dataclass(kw_only=True)
class Gui(LuaDict):
    name: str
    path: str


@dataclass(kw_only=True)
class Asset(LuaDict):
    identifier: str
    parent: str
    gui: Gui
    renderable: Optional[Renderable] = None
    transform: Optional[Transform] = None


def orbits_to_sbdb_file(orbits: Orbits, path: str) -> str:
    # Convert to Keplerian elements in heliocentric ecliptic J2000 frame
    keplerian = transform_coordinates(
        orbits.coordinates,
        representation_out=KeplerianCoordinates,
        frame_out="ecliptic",
        origin_out=OriginCodes.SUN,
    )

    # Write epochs in correct format
    epochs_tdb = orbits.coordinates.time.rescale("tdb").to_astropy()
    epochs = pa.array(
        [
            iso.split(" ")[0] + "." + f"{mjd}".split(".")[1]
            for iso, mjd in zip(epochs_tdb.iso, epochs_tdb.mjd)
        ]
    )

    table = pa.Table.from_pydict(
        {
            "full_name": orbits.orbit_id,
            "epoch_cal": epochs,
            "e": keplerian.e,
            "a": keplerian.a,
            "i": keplerian.i,
            "om": keplerian.raan,
            "w": keplerian.ap,
            "ma": keplerian.M,
            "per": keplerian.P,
        }
    )

    table.to_pandas().to_csv(path, index=False)
    return


def create_initialization(assets: List[str]) -> str:

    initialization = ["asset.onInitialize(function ()"]
    deinitialization = ["asset.onDeinitialize(function ()"]
    for asset in assets:
        initialization.append(f"  openspace.addSceneGraphNode({asset});")
        deinitialization.append(f"  openspace.removeSceneGraphNode({asset});")

    initialization.append("end)")
    deinitialization.append("end)")
    combined = "\n".join(initialization) + "\n" + "\n".join(deinitialization)
    return combined


def create_renderable_orbital_kepler(
    orbits: Orbits,
    out_dir: str,
    identifier: str,
    gui_name: Optional[str] = None,
    gui_path: Optional[str] = None,
    color: Tuple[float, float, float] = (1.0, 1.0, 1.0),
    segment_quality: int = 10,
    contiguous_mode: Optional[bool] = None,
    enable_max_size: Optional[bool] = None,
    enable_outline: Optional[bool] = None,
    max_size: Optional[float] = None,
    outline_color: Optional[Tuple[float, float, float]] = None,
    outline_width: Optional[float] = None,
    point_size_exponent: Optional[float] = None,
    rendering: Optional[Literal["Trail", "Point", "PointsTrails"]] = None,
    render_size: Optional[int] = None,
    start_render_idx: Optional[int] = None,
    trail_fade: Optional[float] = None,
    trail_width: Optional[float] = None,
    dim_in_atmosphere: Optional[bool] = None,
    enabled: Optional[bool] = None,
    opacity: Optional[float] = None,
    render_bin_mode: Literal["Opaque", "Transparent", "Both"] = None,
    tag: Optional[Union[str, List[str]]] = None,
):
    """
    Create a renderable orbital Kepler for a given set of orbits.

    This is the best renderable to use for large numbers of orbits (e.g. > 1000).

    Example
    -------
    >>> from adam_core.orbits.query import query_sbdb
    >>> from adam_core.orbits.openspace import create_renderable_orbital_kepler
    >>> orbits = query_sbdb(["2013 RR165", "2018 BP1"])
    >>> create_renderable_orbital_kepler(orbits, "out_dir", "openspace_example")

    Parameters
    ----------
    orbits : Orbits
        The orbits to create a renderable orbitals for.
    out_dir : str
        The directory to output the asset files
    identifier : str
        The identifier for the asset
    gui_name : str, optional
        The name of the GUI for the asset.
    gui_path : str, optional
        The path of the GUI for the asset.
    color : tuple, optional
        The color of the Keplerian orbital.
    segment_quality : int, optional
        The segment quality for the Keplerian orbital.
    contiguous_mode : bool, optional
        Whether to enable contiguous mode for the Keplerian orbital.
    enable_max_size : bool, optional
        Whether to enable max size for the Keplerian orbital.
    enable_outline : bool, optional
        Whether to enable outline for the Keplerian orbital.
    max_size : float, optional
        The max size for the Keplerian orbital.
    outline_color : tuple, optional
        The color of the outline for the Keplerian orbital.
    outline_width : float, optional
        The width of the outline for the Keplerian orbital.
    point_size_exponent : float, optional
        The size exponent for the Keplerian orbital.
    rendering : str, optional
        The rendering type for the Keplerian orbital.
    render_size : int, optional
        The render size for the Keplerian orbital.
    start_render_idx : int, optional
        The start render index for the Keplerian orbital.
    trail_fade : float, optional
        The trail fade for the Keplerian orbital.
    trail_width : float, optional
        The trail width for the Keplerian orbital.
    dim_in_atmosphere : bool, optional
        Whether to render the Keplerian orbital in the atmosphere.
    enabled : bool, optional
        Whether to enable the Keplerian orbital.
    opacity : float, optional
        The opacity of the Keplerian orbital.
    render_bin_mode : str, optional
        The render bin mode for the Keplerian orbital.
    tag : str, optional
        The tag for the Keplerian orbital.
    """
    safe_identifier = identifier.replace(" ", "_")

    if rendering is not None:
        rendering = RenderableOrbitalKeplerRendering(rendering)

    if render_bin_mode is not None:
        render_bin_mode = RenderBinMode(render_bin_mode)

    if gui_name is None:
        gui_name = safe_identifier

    if gui_path is None:
        gui_path = "/ADAM"

    gui = Gui(name=gui_name, path=gui_path)

    # Create SBDB formatted file
    os.makedirs(out_dir, exist_ok=True)
    path = os.path.join(out_dir, f"{safe_identifier}.csv")
    orbits_to_sbdb_file(orbits, path)

    # Initialize the renderable
    renderable = RenderableOrbitalKepler(
        color=color,
        format=RenderableOrbitalKeplerFormat.SBDB,
        path=Resource(path=os.path.basename(path)),
        segment_quality=segment_quality,
        contiguous_mode=contiguous_mode,
        enable_max_size=enable_max_size,
        enable_outline=enable_outline,
        max_size=max_size,
        outline_color=outline_color,
        outline_width=outline_width,
        point_size_exponent=point_size_exponent,
        rendering=rendering,
        render_size=render_size,
        start_render_idx=start_render_idx,
        trail_fade=trail_fade,
        trail_width=trail_width,
        dim_in_atmosphere=dim_in_atmosphere,
        enabled=enabled,
        opacity=opacity,
        render_bin_mode=render_bin_mode,
        tag=tag,
    )

    # Declare the asset
    asset = Asset(
        identifier=safe_identifier,
        parent="SunEclipJ2000",
        renderable=renderable,
        gui=gui,
    )

    with open(os.path.join(out_dir, f"{safe_identifier}.asset"), "w") as f:
        f.write("local Object = ")
        f.write(asset.to_string(indent=4))
        f.write("\n\n")
        f.write(create_initialization(["Object"]))

    return


def create_renderable_trail_orbit(
    orbits,
    out_dir: str,
    identifier: str,
    trail_head: Optional[bool] = True,
    gui_name: Optional[str] = None,
    gui_path: Optional[str] = None,
    color: Tuple[float, float, float] = (1.0, 1.0, 1.0),
    resolution: int = 86400,
    translation_type: Literal["Kepler", "Spice"] = "Kepler",
    enable_fade: Optional[bool] = None,
    line_fade_amount: Optional[float] = None,
    line_length: Optional[float] = None,
    line_width: Optional[float] = None,
    point_size: Optional[int] = None,
    rendering: Literal["Lines", "Points", "Lines+Points"] = None,
    dim_in_atmosphere: Optional[bool] = None,
    enabled: Optional[bool] = None,
    opacity: Optional[float] = None,
    period: Optional[float] = None,
    render_bin_mode: Literal["Opaque", "Transparent", "Both"] = None,
    tag: Optional[Union[str, List[str]]] = None,
    spice_kernel_path: Optional[str] = None,
    spice_id_mappings: Optional[dict] = None,
):
    """
    Create a renderable trail orbit for a given set of orbits. These orbits can be represented by
    two different "translation" types:

    - KeplerTranslation: This is a Keplerian translation, which takes in a set of Keplerian elements and a time (default)
    - SpiceTranslation: This is a translation that is based on SPICE kernel data. SpiceKernels need to be created (see. )
    and passed to this function.

    The Keplerian translation is the default and is the best for small numbers of orbits (e.g. < 1000).
    The Spice translation is the best to use for a few dozen orbits.

    Example
    -------
    >>> from adam_core.orbits.query import query_sbdb
    >>> from adam_core.orbits.openspace import create_renderable_trail_orbit
    >>> orbits = query_sbdb(["2013 RR165", "2018 BP1"])
    >>> create_renderable_trail_orbit(orbits, "out_dir", "openspace_example", translation_type="Kepler")

    Parameters
    ----------
    orbits : Orbits
        The orbits to create a renderable trail orbit for
    out_dir : str
        The directory to output the asset files
    identifier : str
        The identifier for the asset
    trail_head : bool, optional
        Whether to include a trail head in the asset.
    gui_name : str, optional
        The name of the GUI for the asset.
    gui_path : str, optional
        The path of the GUI for the asset.
    color : tuple, optional
        The color of the trail.
    resolution : int, optional
        The resolution of the trail.
    translation_type : str, optional
        The type of translation to use.
    enable_fade : bool, optional
        Whether to enable fade for the trail.
    line_fade_amount : float, optional
        The amount of fade for the trail.
    line_length : float, optional
        The length of the trail.
    line_width : float, optional
        The width of the trail.
    point_size : int, optional
        The size of the points in the trail.
    rendering : str, optional
        The rendering type for the trail.
    dim_in_atmosphere : bool, optional
        Whether to render the trail in the atmosphere.
    enabled : bool, optional
        Whether to enable the trail.
    opacity : float, optional
        The opacity of the trail.
    period : float, optional
        The period of the trail.
    render_bin_mode : str, optional
        The render bin mode for the trail.
    tag : str, optional
        The tag for the trail.
    spice_kernel_path : str, optional
        The path to the SPICE kernel file.
    spice_id_mappings : dict, optional
        A dictionary mapping object IDs to SPICE IDs.
    """
    safe_identifier = identifier.replace(" ", "_")

    if rendering is not None:
        rendering = RenderableTrailRendering(rendering)

    if render_bin_mode is not None:
        render_bin_mode = RenderBinMode(render_bin_mode)

    if gui_path is None:
        gui_path = "/ADAM"

    keplerian = transform_coordinates(
        orbits.coordinates,
        representation_out=KeplerianCoordinates,
        frame_out="ecliptic",
        origin_out=OriginCodes.SUN,
    )

    if translation_type == "Spice":
        if spice_kernel_path is None:
            raise ValueError("Spice kernel path is required for Spice translation")
        if spice_id_mappings is None:
            raise ValueError("Spice ID mappings are required for Spice translation")

    os.makedirs(out_dir, exist_ok=True)

    with open(os.path.join(out_dir, f"{safe_identifier}.asset"), "w") as f:

        assets = []
        asset_number = 0
        for i, (orbit, keplerian_i) in enumerate(zip(orbits, keplerian)):

            orbit_id = orbit.orbit_id[0].as_py()
            object_id = orbit.object_id[0].as_py()
            if object_id is None:
                object_id = orbit_id

            if trail_head:
                gui_trail = Gui(
                    name=f"{object_id} Trail", path=gui_path + f"/{object_id}"
                )
                gui_head = Gui(
                    name=f"{object_id} Head", path=gui_path + f"/{object_id}"
                )
            else:
                gui_trail = Gui(name=f"{object_id}", path=gui_path)

            if translation_type == "Kepler":
                translation = KeplerTranslation(
                    epoch=keplerian_i.time[0].rescale("tdb").to_astropy().iso[0],
                    semi_major_axis=keplerian_i.a[0].as_py() * KM_P_AU,
                    eccentricity=keplerian_i.e[0].as_py(),
                    inclination=keplerian_i.i[0].as_py(),
                    argument_of_periapsis=keplerian_i.ap[0].as_py(),
                    ascending_node=keplerian_i.raan[0].as_py(),
                    mean_anomaly=keplerian_i.M[0].as_py(),
                    period=_safe_orbital_period(keplerian_i.P[0])
                    * S_P_DAY,  # in seconds here
                )
            elif translation_type == "Spice":
                translation = SpiceTranslation(
                    target=spice_id_mappings[object_id],
                    observer="SOLAR SYSTEM BARYCENTER",
                )
            else:
                raise ValueError(f"Invalid translation type: {translation}")

            renderable = RenderableTrailOrbit(
                color=color,
                period=(
                    _safe_orbital_period(keplerian_i.P[0]) if period is None else period
                ),  # in days here (confusingly)
                resolution=resolution,
                translation=translation,
                enable_fade=enable_fade,
                line_fade_amount=line_fade_amount,
                line_length=line_length,
                line_width=line_width,
                point_size=point_size,
                rendering=rendering,
            )

            asset = Asset(
                identifier=object_id.replace(" ", "_"),
                parent="SunEclipJ2000",
                renderable=renderable,
                gui=gui_trail,
            )

            asset_id = f"Object{asset_number:08d}"
            asset_number += 1
            assets.append(asset_id)

            f.write(f"local {asset_id} = ")
            f.write(asset.to_string(indent=4))
            f.write("\n\n")

            if trail_head:
                head_asset = Asset(
                    identifier=object_id.replace(" ", "_") + "_Head",
                    parent="SunEclipJ2000",
                    transform=Transform(
                        translation=translation,
                    ),
                    gui=gui_head,
                )
                head_asset_id = f"Object{asset_number:08d}"
                asset_number += 1
                assets.append(head_asset_id)

                f.write(f"local {head_asset_id} = ")
                f.write(head_asset.to_string(indent=4))
                f.write("\n\n")

    initialization = create_initialization(assets)
    with open(os.path.join(out_dir, f"{safe_identifier}.asset"), "a") as f:
        f.write(initialization)

        if translation_type == "Spice":
            f.write("asset.onInitialize(function()\n")
            f.write(
                f"  local kernelResource = openspace.resource('{os.path.relpath(spice_kernel_path, out_dir)}')\n"
            )
            f.write("  openspace.spice.loadKernel(kernelResource)\n")
            f.write("end)\n")
            f.write("asset.onDeinitialize(function()\n")
            f.write("  openspace.spice.unloadKernel(kernelResource)\n")
            f.write("end)\n")

    return
